from pathlib import Path
import logging as _logging

from PyQt5.QtWidgets import QSizePolicy
from PyQt5.QtWebEngineWidgets import QWebEngineView

from ..app.graph import (generate_graph, sutta_nodes_and_edges,
                         dict_word_nodes_and_edges,
                         document_page_nodes_and_edges, sutta_graph_id)

from ..app.file_doc import FileDoc
from ..app.db import userdata_models as Um

from ..app.types import USutta, UDictWord

logger = _logging.getLogger(__name__)


class HasLinksSidebar:
    def init_links_sidebar(self):
        self.setup_content_graph()

    def setup_content_graph(self):
        self.content_graph = QWebEngineView()
        self.content_graph.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.content_graph.setHtml('')
        self.content_graph.show()
        self.links_layout.addWidget(self.content_graph)

    def generate_graph_for_sutta(self,
                                 sutta: USutta,
                                 queue_id: str,
                                 graph_path: Path,
                                 messages_url: str):

        (nodes, edges) = sutta_nodes_and_edges(app_data=self._app_data, sutta=sutta, distance=3)

        hits = len(nodes) - 1
        if hits > 0:
            self.rightside_tabs.setTabText(self.links_tab_idx, f"Links ({hits})")
        else:
            self.rightside_tabs.setTabText(self.links_tab_idx, "Links")

        selected = []

        for idx, n in enumerate(nodes):
            if n[0] == sutta_graph_id(sutta):
                selected.append(idx)

        generate_graph(nodes, edges, selected, queue_id, graph_path, messages_url)

    def generate_graph_for_dict_word(self,
                                     dict_word: UDictWord,
                                     queue_id: str,
                                     graph_path: Path,
                                     messages_url: str):

        (nodes, edges) = dict_word_nodes_and_edges(app_data=self._app_data, dict_word=dict_word, distance=3)

        hits = len(nodes) - 1
        if hits > 0:
            self.rightside_tabs.setTabText(self.links_tab_idx, f"Links ({hits})")
        else:
            self.rightside_tabs.setTabText(self.links_tab_idx, "Links")

        # central node was appended last
        selected = [len(nodes) - 1]

        generate_graph(nodes, edges, selected, queue_id, graph_path, messages_url)

    def generate_graph_for_document(self,
                                    file_doc: FileDoc,
                                    db_doc: Um.Document,
                                    queue_id: str,
                                    graph_path: Path,
                                    messages_url: str):

        (nodes, edges) = document_page_nodes_and_edges(
            app_data=self._app_data,
            db_doc=db_doc,
            page_number=file_doc._current_idx + 1,
            distance=3
        )

        hits = len(nodes) - 1
        if hits > 0:
            self.rightside_tabs.setTabText(self.links_tab_idx, f"Links ({hits})")
        else:
            self.rightside_tabs.setTabText(self.links_tab_idx, "Links")

        # central node was appended last
        selected = [len(nodes) - 1]

        generate_graph(nodes, edges, selected, queue_id, graph_path, messages_url)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'simsapa/assets/ui/dictionary_search_window.ui'
#
# Created by: PyQt5 UI code generator 5.15.6
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_DictionarySearchWindow(object):
    def setupUi(self, DictionarySearchWindow):
        DictionarySearchWindow.setObjectName("DictionarySearchWindow")
        DictionarySearchWindow.resize(1068, 642)
        DictionarySearchWindow.setBaseSize(QtCore.QSize(800, 600))
        self.central_widget = QtWidgets.QWidget(DictionarySearchWindow)
        self.central_widget.setObjectName("central_widget")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout(self.central_widget)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.main_layout = QtWidgets.QVBoxLayout()
        self.main_layout.setObjectName("main_layout")
        self.searchbar_layout = QtWidgets.QHBoxLayout()
        self.searchbar_layout.setSizeConstraint(QtWidgets.QLayout.SetMinimumSize)
        self.searchbar_layout.setObjectName("searchbar_layout")
        self.search_input = QtWidgets.QLineEdit(self.central_widget)
        self.search_input.setMinimumSize(QtCore.QSize(0, 35))
        self.search_input.setClearButtonEnabled(True)
        self.search_input.setObjectName("search_input")
        self.searchbar_layout.addWidget(self.search_input)
        self.search_button = QtWidgets.QPushButton(self.central_widget)
        self.search_button.setMinimumSize(QtCore.QSize(0, 40))
        self.search_button.setObjectName("search_button")
        self.searchbar_layout.addWidget(self.search_button)
        self.main_layout.addLayout(self.searchbar_layout)
        self.splitter = QtWidgets.QSplitter(self.central_widget)
        self.splitter.setOrientation(QtCore.Qt.Horizontal)
        self.splitter.setObjectName("splitter")
        self.verticalLayoutWidget_2 = QtWidgets.QWidget(self.splitter)
        self.verticalLayoutWidget_2.setObjectName("verticalLayoutWidget_2")
        self.content_layout = QtWidgets.QVBoxLayout(self.verticalLayoutWidget_2)
        self.content_layout.setContentsMargins(0, 0, 0, 0)
        self.content_layout.setObjectName("content_layout")
        spacerItem = QtWidgets.QSpacerItem(500, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.content_layout.addItem(spacerItem)
        self.verticalLayoutWidget_3 = QtWidgets.QWidget(self.splitter)
        self.verticalLayoutWidget_3.setObjectName("verticalLayoutWidget_3")
        self.tabs_layout = QtWidgets.QVBoxLayout(self.verticalLayoutWidget_3)
        self.tabs_layout.setContentsMargins(0, 0, 0, 0)
        self.tabs_layout.setObjectName("tabs_layout")
        self.rightside_tabs = QtWidgets.QTabWidget(self.verticalLayoutWidget_3)
        self.rightside_tabs.setMinimumSize(QtCore.QSize(500, 500))
        self.rightside_tabs.setObjectName("rightside_tabs")
        self.results_tab = QtWidgets.QWidget()
        self.results_tab.setObjectName("results_tab")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(self.results_tab)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.results_pages_layout = QtWidgets.QHBoxLayout()
        self.results_pages_layout.setObjectName("results_pages_layout")
        self.results_page_input = QtWidgets.QSpinBox(self.results_tab)
        self.results_page_input.setMinimum(1)
        self.results_page_input.setMaximum(999)
        self.results_page_input.setObjectName("results_page_input")
        self.results_pages_layout.addWidget(self.results_page_input)
        self.results_label = QtWidgets.QLabel(self.results_tab)
        self.results_label.setObjectName("results_label")
        self.results_pages_layout.addWidget(self.results_label)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.results_pages_layout.addItem(spacerItem1)
        self.results_first_page_btn = QtWidgets.QPushButton(self.results_tab)
        self.results_first_page_btn.setObjectName("results_first_page_btn")
        self.results_pages_layout.addWidget(self.results_first_page_btn)
        self.results_last_page_btn = QtWidgets.QPushButton(self.results_tab)
        self.results_last_page_btn.setObjectName("results_last_page_btn")
        self.results_pages_layout.addWidget(self.results_last_page_btn)
        self.verticalLayout.addLayout(self.results_pages_layout)
        self.results_list = QtWidgets.QListWidget(self.results_tab)
        self.results_list.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.results_list.setObjectName("results_list")
        self.verticalLayout.addWidget(self.results_list)
        self.verticalLayout_4.addLayout(self.verticalLayout)
        self.rightside_tabs.addTab(self.results_tab, "")
        self.links_tab = QtWidgets.QWidget()
        self.links_tab.setObjectName("links_tab")
        self.verticalLayout_5 = QtWidgets.QVBoxLayout(self.links_tab)
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.links_layout = QtWidgets.QVBoxLayout()
        self.links_layout.setObjectName("links_layout")
        self.verticalLayout_5.addLayout(self.links_layout)
        self.rightside_tabs.addTab(self.links_tab, "")
        self.memos_tab = QtWidgets.QWidget()
        self.memos_tab.setObjectName("memos_tab")
        self.verticalLayout_6 = QtWidgets.QVBoxLayout(self.memos_tab)
        self.verticalLayout_6.setObjectName("verticalLayout_6")
        self.memos_layout = QtWidgets.QVBoxLayout()
        self.memos_layout.setObjectName("memos_layout")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.add_memo_button = QtWidgets.QPushButton(self.memos_tab)
        self.add_memo_button.setObjectName("add_memo_button")
        self.horizontalLayout.addWidget(self.add_memo_button)
        self.clear_memo_button = QtWidgets.QPushButton(self.memos_tab)
        self.clear_memo_button.setObjectName("clear_memo_button")
        self.horizontalLayout.addWidget(self.clear_memo_button)
        self.remove_memo_button = QtWidgets.QPushButton(self.memos_tab)
        self.remove_memo_button.setObjectName("remove_memo_button")
        self.horizontalLayout.addWidget(self.remove_memo_button)
        self.memos_layout.addLayout(self.horizontalLayout)
        self.label_2 = QtWidgets.QLabel(self.memos_tab)
        self.label_2.setObjectName("label_2")
        self.memos_layout.addWidget(self.label_2)
        self.front_input = QtWidgets.QPlainTextEdit(self.memos_tab)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.front_input.sizePolicy().hasHeightForWidth())
        self.front_input.setSizePolicy(sizePolicy)
        self.front_input.setMaximumSize(QtCore.QSize(16777215, 50))
        self.front_input.setObjectName("front_input")
        self.memos_layout.addWidget(self.front_input)
        self.label_5 = QtWidgets.QLabel(self.memos_tab)
        self.label_5.setObjectName("label_5")
        self.memos_layout.addWidget(self.label_5)
        self.back_input = QtWidgets.QPlainTextEdit(self.memos_tab)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.back_input.sizePolicy().hasHeightForWidth())
        self.back_input.setSizePolicy(sizePolicy)
        self.back_input.setMaximumSize(QtCore.QSize(16777215, 50))
        self.back_input.setObjectName("back_input")
        self.memos_layout.addWidget(self.back_input)
        self.label_3 = QtWidgets.QLabel(self.memos_tab)
        self.label_3.setObjectName("label_3")
        self.memos_layout.addWidget(self.label_3)
        self.memos_list = QtWidgets.QListView(self.memos_tab)
        self.memos_list.setObjectName("memos_list")
        self.memos_layout.addWidget(self.memos_list)
        self.verticalLayout_6.addLayout(self.memos_layout)
        self.rightside_tabs.addTab(self.memos_tab, "")
        self.recent_tab = QtWidgets.QWidget()
        self.recent_tab.setObjectName("recent_tab")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.recent_tab)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.recent_layout = QtWidgets.QVBoxLayout()
        self.recent_layout.setObjectName("recent_layout")
        self.recent_list = QtWidgets.QListWidget(self.recent_tab)
        self.recent_list.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.recent_list.setObjectName("recent_list")
        self.recent_layout.addWidget(self.recent_list)
        self.verticalLayout_3.addLayout(self.recent_layout)
        self.rightside_tabs.addTab(self.recent_tab, "")
        self.tabs_layout.addWidget(self.rightside_tabs)
        self.main_layout.addWidget(self.splitter)
        self.horizontalLayout_2.addLayout(self.main_layout)
        DictionarySearchWindow.setCentralWidget(self.central_widget)
        self.menubar = QtWidgets.QMenuBar(DictionarySearchWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 1068, 25))
        self.menubar.setObjectName("menubar")
        self.menu_File = QtWidgets.QMenu(self.menubar)
        self.menu_File.setObjectName("menu_File")
        self.menu_Edit = QtWidgets.QMenu(self.menubar)
        self.menu_Edit.setObjectName("menu_Edit")
        self.menu_Windows = QtWidgets.QMenu(self.menubar)
        self.menu_Windows.setObjectName("menu_Windows")
        self.menu_Help = QtWidgets.QMenu(self.menubar)
        self.menu_Help.setObjectName("menu_Help")
        self.menu_Dictionary = QtWidgets.QMenu(self.menubar)
        self.menu_Dictionary.setObjectName("menu_Dictionary")
        DictionarySearchWindow.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(DictionarySearchWindow)
        self.statusbar.setObjectName("statusbar")
        DictionarySearchWindow.setStatusBar(self.statusbar)
        self.toolBar = QtWidgets.QToolBar(DictionarySearchWindow)
        self.toolBar.setObjectName("toolBar")
        DictionarySearchWindow.addToolBar(QtCore.Qt.LeftToolBarArea, self.toolBar)
        self.action_Copy = QtWidgets.QAction(DictionarySearchWindow)
        self.action_Copy.setObjectName("action_Copy")
        self.action_Paste = QtWidgets.QAction(DictionarySearchWindow)
        self.action_Paste.setObjectName("action_Paste")
        self.action_Quit = QtWidgets.QAction(DictionarySearchWindow)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/close"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.action_Quit.setIcon(icon)
        self.action_Quit.setObjectName("action_Quit")
        self.action_Sutta_Search = QtWidgets.QAction(DictionarySearchWindow)
        icon1 = QtGui.QIcon()
        icon1.addPixmap(QtGui.QPixmap(":/book"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.action_Sutta_Search.setIcon(icon1)
        self.action_Sutta_Search.setObjectName("action_Sutta_Search")
        self.action_Dictionary_Search = QtWidgets.QAction(DictionarySearchWindow)
        icon2 = QtGui.QIcon()
        icon2.addPixmap(QtGui.QPixmap(":/dictionary"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.action_Dictionary_Search.setIcon(icon2)
        self.action_Dictionary_Search.setObjectName("action_Dictionary_Search")
        self.action_About = QtWidgets.QAction(DictionarySearchWindow)
        self.action_About.setObjectName("action_About")
        self.action_Website = QtWidgets.QAction(DictionarySearchWindow)
        self.action_Website.setObjectName("action_Website")
        self.action_Close_Window = QtWidgets.QAction(DictionarySearchWindow)
        self.action_Close_Window.setObjectName("action_Close_Window")
        self.action_Open = QtWidgets.QAction(DictionarySearchWindow)
        self.action_Open.setObjectName("action_Open")
        self.action_Document_Reader = QtWidgets.QAction(DictionarySearchWindow)
        self.action_Document_Reader.setObjectName("action_Document_Reader")
        self.action_Library = QtWidgets.QAction(DictionarySearchWindow)
        self.action_Library.setObjectName("action_Library")
        self.action_Memos = QtWidgets.QAction(DictionarySearchWindow)
        self.action_Memos.setObjectName("action_Memos")
        self.action_Dictionaries_Manager = QtWidgets.QAction(DictionarySearchWindow)
        self.action_Dictionaries_Manager.setObjectName("action_Dictionaries_Manager")
        self.action_Links = QtWidgets.QAction(DictionarySearchWindow)
        self.action_Links.setObjectName("action_Links")
        self.action_Import_from_StarDict = QtWidgets.QAction(DictionarySearchWindow)
        self.action_Import_from_StarDict.setObjectName("action_Import_from_StarDict")
        self.menu_File.addAction(self.action_Open)
        self.menu_File.addAction(self.action_Close_Window)
        self.menu_File.addSeparator()
        self.menu_File.addAction(self.action_Quit)
        self.menu_Edit.addAction(self.action_Copy)
        self.menu_Edit.addAction(self.action_Paste)
        self.menu_Windows.addAction(self.action_Sutta_Search)
        self.menu_Windows.addAction(self.action_Dictionary_Search)
        self.menu_Windows.addAction(self.action_Dictionaries_Manager)
        self.menu_Windows.addAction(self.action_Document_Reader)
        self.menu_Windows.addAction(self.action_Library)
        self.menu_Windows.addAction(self.action_Memos)
        self.menu_Windows.addAction(self.action_Links)
        self.menu_Help.addAction(self.action_Website)
        self.menu_Help.addAction(self.action_About)
        self.menu_Dictionary.addAction(self.action_Import_from_StarDict)
        self.menubar.addAction(self.menu_File.menuAction())
        self.menubar.addAction(self.menu_Edit.menuAction())
        self.menubar.addAction(self.menu_Windows.menuAction())
        self.menubar.addAction(self.menu_Dictionary.menuAction())
        self.menubar.addAction(self.menu_Help.menuAction())
        self.toolBar.addAction(self.action_Sutta_Search)
        self.toolBar.addAction(self.action_Dictionary_Search)

        self.retranslateUi(DictionarySearchWindow)
        self.rightside_tabs.setCurrentIndex(0)
        QtCore.QMetaObject.connectSlotsByName(DictionarySearchWindow)

    def retranslateUi(self, DictionarySearchWindow):
        _translate = QtCore.QCoreApplication.translate
        DictionarySearchWindow.setWindowTitle(_translate("DictionarySearchWindow", "Dictionary Search - Simsapa"))
        self.search_button.setText(_translate("DictionarySearchWindow", "Search"))
        self.results_label.setText(_translate("DictionarySearchWindow", "Showing a-b out of x"))
        self.results_first_page_btn.setText(_translate("DictionarySearchWindow", "First"))
        self.results_last_page_btn.setText(_translate("DictionarySearchWindow", "Last"))
        self.rightside_tabs.setTabText(self.rightside_tabs.indexOf(self.results_tab), _translate("DictionarySearchWindow", "Results"))
        self.rightside_tabs.setTabText(self.rightside_tabs.indexOf(self.links_tab), _translate("DictionarySearchWindow", "Links"))
        self.add_memo_button.setText(_translate("DictionarySearchWindow", "Add"))
        self.clear_memo_button.setText(_translate("DictionarySearchWindow", "Clear"))
        self.remove_memo_button.setText(_translate("DictionarySearchWindow", "Remove"))
        self.label_2.setText(_translate("DictionarySearchWindow", "Front"))
        self.label_5.setText(_translate("DictionarySearchWindow", "Back"))
        self.label_3.setText(_translate("DictionarySearchWindow", "Memos for this page"))
        self.rightside_tabs.setTabText(self.rightside_tabs.indexOf(self.memos_tab), _translate("DictionarySearchWindow", "Memos"))
        self.rightside_tabs.setTabText(self.rightside_tabs.indexOf(self.recent_tab), _translate("DictionarySearchWindow", "Recent"))
        self.menu_File.setTitle(_translate("DictionarySearchWindow", "&File"))
        self.menu_Edit.setTitle(_translate("DictionarySearchWindow", "&Edit"))
        self.menu_Windows.setTitle(_translate("DictionarySearchWindow", "&Windows"))
        self.menu_Help.setTitle(_translate("DictionarySearchWindow", "&Help"))
        self.menu_Dictionary.setTitle(_translate("DictionarySearchWindow", "&Dictionary"))
        self.toolBar.setWindowTitle(_translate("DictionarySearchWindow", "toolBar"))
        self.action_Copy.setText(_translate("DictionarySearchWindow", "&Copy"))
        self.action_Paste.setText(_translate("DictionarySearchWindow", "&Paste"))
        self.action_Quit.setText(_translate("DictionarySearchWindow", "&Quit"))
        self.action_Quit.setShortcut(_translate("DictionarySearchWindow", "Ctrl+Q"))
        self.action_Sutta_Search.setText(_translate("DictionarySearchWindow", "&Sutta Search"))
        self.action_Sutta_Search.setShortcut(_translate("DictionarySearchWindow", "F5"))
        self.action_Dictionary_Search.setText(_translate("DictionarySearchWindow", "&Dictionary Search"))
        self.action_Dictionary_Search.setShortcut(_translate("DictionarySearchWindow", "F6"))
        self.action_About.setText(_translate("DictionarySearchWindow", "&About"))
        self.action_Website.setText(_translate("DictionarySearchWindow", "&Website"))
        self.action_Close_Window.setText(_translate("DictionarySearchWindow", "&Close Window"))
        self.action_Open.setText(_translate("DictionarySearchWindow", "&Open..."))
        self.action_Open.setShortcut(_translate("DictionarySearchWindow", "Ctrl+O"))
        self.action_Document_Reader.setText(_translate("DictionarySearchWindow", "D&ocument Reader"))
        self.action_Document_Reader.setToolTip(_translate("DictionarySearchWindow", "Document Reader"))
        self.action_Document_Reader.setShortcut(_translate("DictionarySearchWindow", "F7"))
        self.action_Library.setText(_translate("DictionarySearchWindow", "&Library"))
        self.action_Library.setShortcut(_translate("DictionarySearchWindow", "F8"))
        self.action_Memos.setText(_translate("DictionarySearchWindow", "&Memos"))
        self.action_Memos.setToolTip(_translate("DictionarySearchWindow", "Memos"))
        self.action_Memos.setShortcut(_translate("DictionarySearchWindow", "F9"))
        self.action_Dictionaries_Manager.setText(_translate("DictionarySearchWindow", "Dictionaries &Manager"))
        self.action_Dictionaries_Manager.setShortcut(_translate("DictionarySearchWindow", "F10"))
        self.action_Links.setText(_translate("DictionarySearchWindow", "&Links"))
        self.action_Links.setShortcut(_translate("DictionarySearchWindow", "F11"))
        self.action_Import_from_StarDict.setText(_translate("DictionarySearchWindow", "&Import from StarDict..."))
from simsapa.assets import icons_rc



import logging as _logging
from pathlib import Path
import requests

import re
import bleach

from sqlalchemy import create_engine
from sqlalchemy.engine import Engine
from sqlalchemy_utils import database_exists, create_database
from alembic import command
from alembic.config import Config
from alembic.script import ScriptDirectory
from alembic.runtime.migration import MigrationContext

from .db import appdata_models as Am
from .db import userdata_models as Um

from simsapa import ALEMBIC_INI, ALEMBIC_DIR

logger = _logging.getLogger(__name__)

def download_file(url: str, folder_path: Path) -> Path:
    file_name = url.split('/')[-1]
    file_path = folder_path.joinpath(file_name)

    with requests.get(url, stream=True) as r:
        r.raise_for_status()
        with open(file_path, 'wb') as f:
            for chunk in r.iter_content(chunk_size=8192):
                f.write(chunk)

    return file_path

def compactPlainText(text: str) -> str:
    # NOTE: Don't remove new lines here, useful for matching beginning of lines when setting snippets.
    # Replace multiple spaces to one.
    text = re.sub(r"  +", ' ', text)

    return text

def compactRichText(text: str) -> str:
    text = bleach.clean(text, tags=[], styles=[], strip=True)
    text = compactPlainText(text)

    return text

def find_or_create_db(db_path: Path, schema_name: str):
    # Create an in-memory database
    engine = create_engine("sqlite+pysqlite://", echo=False)

    if isinstance(engine, Engine):
        db_conn = engine.connect()
        db_url = f"sqlite+pysqlite:///{db_path}"

        alembic_cfg = Config(f"{ALEMBIC_INI}")
        alembic_cfg.set_main_option('script_location', f"{ALEMBIC_DIR}")
        alembic_cfg.set_main_option('sqlalchemy.url', db_url)

        if not database_exists(db_url):
            logger.info(f"Cannot find {db_url}, creating it")
            # On a new install, create database and all tables with the recent schema.
            create_database(db_url)
            db_conn.execute(f"ATTACH DATABASE '{db_path}' AS '{schema_name}';")
            if schema_name == 'userdata':
                Um.metadata.create_all(bind=engine)
            else:
                Am.metadata.create_all(bind=engine)

            # generate the Alembic version table, "stamping" it with the most recent rev:
            command.stamp(alembic_cfg, "head")

        elif not is_db_revision_at_head(alembic_cfg, engine):
            logger.info(f"{db_url} is stale, running migrations")

            if db_conn is not None:
                alembic_cfg.attributes['connection'] = db_conn
                try:
                    command.upgrade(alembic_cfg, "head")
                except Exception as e:
                    # NOTE: logger.error() is not printed for some reason.
                    print("ERROR - Failed to run migrations.")
                    print(e)
                    exit(1)
    else:
        logger.error("Can't create in-memory database")

def is_db_revision_at_head(alembic_cfg: Config, e: Engine) -> bool:
    directory = ScriptDirectory.from_config(alembic_cfg)
    with e.begin() as db_conn:
        context = MigrationContext.configure(db_conn)
        return set(context.get_current_heads()) == set(directory.get_heads())



import os
from functools import partial
from typing import List

from PyQt5.QtWidgets import (QApplication, QMainWindow, QFileDialog)

from .types import AppData

from ..layouts.sutta_search import SuttaSearchWindow
from ..layouts.dictionary_search import DictionarySearchWindow
# from ..layouts.dictionaries_manager import DictionariesManagerWindow
from ..layouts.document_reader import DocumentReaderWindow
from ..layouts.library_browser import LibraryBrowserWindow
from ..layouts.memos_browser import MemosBrowserWindow
from ..layouts.links_browser import LinksBrowserWindow


class AppWindows:
    def __init__(self, app: QApplication, app_data: AppData):
        self._app = app
        self._app_data = app_data
        self._windows: List[QMainWindow] = []

    def _new_sutta_search_window(self):
        view = SuttaSearchWindow(self._app_data)
        self._connect_signals(view)
        view.show()
        if self._app_data.sutta_to_open:
            view._show_sutta(self._app_data.sutta_to_open)
            self._app_data.sutta_to_open = None
        self._windows.append(view)

    def _new_dictionary_search_window(self):
        view = DictionarySearchWindow(self._app_data)
        self._connect_signals(view)
        view.show()
        if self._app_data.dict_word_to_open:
            view._show_word(self._app_data.dict_word_to_open)
            self._app_data.dict_word_to_open = None
        self._windows.append(view)

#    def _new_dictionaries_manager_window(self):
#        view = DictionariesManagerWindow(self._app_data)
#        self._connect_signals(view)
#        view.show()
#        self._windows.append(view)

    def _new_library_browser_window(self):
        view = LibraryBrowserWindow(self._app_data)
        self._connect_signals(view)
        view.show()
        self._windows.append(view)

    def _new_memos_browser_window(self):
        view = MemosBrowserWindow(self._app_data)
        self._connect_signals(view)
        view.show()
        self._windows.append(view)

    def _new_links_browser_window(self):
        view = LinksBrowserWindow(self._app_data)
        self._connect_signals(view)
        view.show()
        self._windows.append(view)

    def _new_document_reader_window(self, file_path=None):
        view = DocumentReaderWindow(self._app_data)
        self._connect_signals(view)
        view.show()

        if file_path is not None and file_path is not False and len(file_path) > 0:
            view.open_doc(file_path)

        self._windows.append(view)

    def _open_selected_document(self, view: QMainWindow):
        doc = view.get_selected_document()
        if doc:
            self._new_document_reader_window(doc.filepath)

    def _open_file_dialog(self, view: QMainWindow):
        try:
            view.open_file_dialog()
        except AttributeError:
            file_path, _ = QFileDialog.getOpenFileName(
                None,
                "Open File...",
                "",
                "PDF or Epub Files (*.pdf *.epub)")

            if len(file_path) != 0:
                self._new_document_reader_window(file_path)

    def _close_all_windows(self):
        for w in self._windows:
            w.close()

    def _quit_app(self):
        self._close_all_windows()
        self._app.quit()

    def _connect_signals(self, view: QMainWindow):
        view.action_Open \
            .triggered.connect(partial(self._open_file_dialog, view))

        try:
            view.action_Open_Selected \
                .triggered.connect(partial(self._open_selected_document, view))
        except Exception:
            # FIXME silent exception
            pass

        view.action_Quit \
            .triggered.connect(partial(self._quit_app))
        view.action_Sutta_Search \
            .triggered.connect(partial(self._new_sutta_search_window))
        view.action_Dictionary_Search \
            .triggered.connect(partial(self._new_dictionary_search_window))
        view.action_Memos \
            .triggered.connect(partial(self._new_memos_browser_window))
        view.action_Links \
            .triggered.connect(partial(self._new_links_browser_window))

        s = os.getenv('ENABLE_WIP_FEATURES')
        if s is not None and s.lower() == 'true':
            # view.action_Dictionaries_Manager \
            #     .triggered.connect(partial(self._new_dictionaries_manager_window))
            view.action_Document_Reader \
                .triggered.connect(partial(self._new_document_reader_window))
            view.action_Library \
                .triggered.connect(partial(self._new_library_browser_window))
        else:
            if hasattr(view,'toolBar'):
                view.toolBar.setVisible(False)

            view.action_Dictionaries_Manager.setVisible(False)
            view.action_Document_Reader.setVisible(False)
            view.action_Library.setVisible(False)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'simsapa/assets/ui/links_browser_window.ui'
#
# Created by: PyQt5 UI code generator 5.15.4
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_LinksBrowserWindow(object):
    def setupUi(self, LinksBrowserWindow):
        LinksBrowserWindow.setObjectName("LinksBrowserWindow")
        LinksBrowserWindow.resize(1068, 643)
        LinksBrowserWindow.setBaseSize(QtCore.QSize(800, 600))
        self.central_widget = QtWidgets.QWidget(LinksBrowserWindow)
        self.central_widget.setObjectName("central_widget")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout(self.central_widget)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.main_layout = QtWidgets.QVBoxLayout()
        self.main_layout.setObjectName("main_layout")
        self.splitter = QtWidgets.QSplitter(self.central_widget)
        self.splitter.setOrientation(QtCore.Qt.Horizontal)
        self.splitter.setObjectName("splitter")
        self.verticalLayoutWidget_2 = QtWidgets.QWidget(self.splitter)
        self.verticalLayoutWidget_2.setObjectName("verticalLayoutWidget_2")
        self.links_layout = QtWidgets.QVBoxLayout(self.verticalLayoutWidget_2)
        self.links_layout.setContentsMargins(0, 0, 0, 0)
        self.links_layout.setObjectName("links_layout")
        spacerItem = QtWidgets.QSpacerItem(500, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.links_layout.addItem(spacerItem)
        self.verticalLayoutWidget_3 = QtWidgets.QWidget(self.splitter)
        self.verticalLayoutWidget_3.setObjectName("verticalLayoutWidget_3")
        self.tabs_layout = QtWidgets.QVBoxLayout(self.verticalLayoutWidget_3)
        self.tabs_layout.setContentsMargins(0, 0, 0, 0)
        self.tabs_layout.setObjectName("tabs_layout")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.create_link_btn = QtWidgets.QPushButton(self.verticalLayoutWidget_3)
        self.create_link_btn.setObjectName("create_link_btn")
        self.horizontalLayout.addWidget(self.create_link_btn)
        self.clear_link_btn = QtWidgets.QPushButton(self.verticalLayoutWidget_3)
        self.clear_link_btn.setObjectName("clear_link_btn")
        self.horizontalLayout.addWidget(self.clear_link_btn)
        self.remove_link_btn = QtWidgets.QPushButton(self.verticalLayoutWidget_3)
        self.remove_link_btn.setObjectName("remove_link_btn")
        self.horizontalLayout.addWidget(self.remove_link_btn)
        self.tabs_layout.addLayout(self.horizontalLayout)
        self.horizontalLayout_5 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_5.setObjectName("horizontalLayout_5")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout()
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.set_from_btn = QtWidgets.QPushButton(self.verticalLayoutWidget_3)
        self.set_from_btn.setObjectName("set_from_btn")
        self.verticalLayout_3.addWidget(self.set_from_btn)
        self.from_view = QtWidgets.QPlainTextEdit(self.verticalLayoutWidget_3)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.from_view.sizePolicy().hasHeightForWidth())
        self.from_view.setSizePolicy(sizePolicy)
        self.from_view.setMaximumSize(QtCore.QSize(16777215, 80))
        self.from_view.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.from_view.setObjectName("from_view")
        self.verticalLayout_3.addWidget(self.from_view)
        self.horizontalLayout_5.addLayout(self.verticalLayout_3)
        self.verticalLayout_2 = QtWidgets.QVBoxLayout()
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.set_to_btn = QtWidgets.QPushButton(self.verticalLayoutWidget_3)
        self.set_to_btn.setObjectName("set_to_btn")
        self.verticalLayout_2.addWidget(self.set_to_btn)
        self.to_view = QtWidgets.QPlainTextEdit(self.verticalLayoutWidget_3)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.to_view.sizePolicy().hasHeightForWidth())
        self.to_view.setSizePolicy(sizePolicy)
        self.to_view.setMaximumSize(QtCore.QSize(16777215, 80))
        self.to_view.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.to_view.setObjectName("to_view")
        self.verticalLayout_2.addWidget(self.to_view)
        self.horizontalLayout_5.addLayout(self.verticalLayout_2)
        self.tabs_layout.addLayout(self.horizontalLayout_5)
        self.searchbar_layout = QtWidgets.QVBoxLayout()
        self.searchbar_layout.setObjectName("searchbar_layout")
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.search_input = QtWidgets.QLineEdit(self.verticalLayoutWidget_3)
        self.search_input.setMinimumSize(QtCore.QSize(0, 35))
        self.search_input.setObjectName("search_input")
        self.horizontalLayout_4.addWidget(self.search_input)
        self.search_button = QtWidgets.QPushButton(self.verticalLayoutWidget_3)
        self.search_button.setObjectName("search_button")
        self.horizontalLayout_4.addWidget(self.search_button)
        self.link_table = QtWidgets.QComboBox(self.verticalLayoutWidget_3)
        self.link_table.setObjectName("link_table")
        self.link_table.addItem("")
        self.link_table.addItem("")
        self.link_table.addItem("")
        self.horizontalLayout_4.addWidget(self.link_table)
        self.searchbar_layout.addLayout(self.horizontalLayout_4)
        self.horizontalLayout_6 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_6.setObjectName("horizontalLayout_6")
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_6.addItem(spacerItem1)
        self.pali_buttons_layout = QtWidgets.QVBoxLayout()
        self.pali_buttons_layout.setObjectName("pali_buttons_layout")
        self.horizontalLayout_6.addLayout(self.pali_buttons_layout)
        spacerItem2 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_6.addItem(spacerItem2)
        self.searchbar_layout.addLayout(self.horizontalLayout_6)
        self.tabs_layout.addLayout(self.searchbar_layout)
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.set_page_number = QtWidgets.QCheckBox(self.verticalLayoutWidget_3)
        self.set_page_number.setEnabled(False)
        self.set_page_number.setObjectName("set_page_number")
        self.horizontalLayout_3.addWidget(self.set_page_number)
        self.page_number = QtWidgets.QSpinBox(self.verticalLayoutWidget_3)
        self.page_number.setEnabled(False)
        self.page_number.setObjectName("page_number")
        self.horizontalLayout_3.addWidget(self.page_number)
        self.tabs_layout.addLayout(self.horizontalLayout_3)
        self.results_list = QtWidgets.QListWidget(self.verticalLayoutWidget_3)
        self.results_list.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.results_list.setLineWidth(1)
        self.results_list.setObjectName("results_list")
        self.tabs_layout.addWidget(self.results_list)
        self.main_layout.addWidget(self.splitter)
        self.horizontalLayout_2.addLayout(self.main_layout)
        LinksBrowserWindow.setCentralWidget(self.central_widget)
        self.menubar = QtWidgets.QMenuBar(LinksBrowserWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 1068, 20))
        self.menubar.setObjectName("menubar")
        self.menu_File = QtWidgets.QMenu(self.menubar)
        self.menu_File.setObjectName("menu_File")
        self.menu_Edit = QtWidgets.QMenu(self.menubar)
        self.menu_Edit.setObjectName("menu_Edit")
        self.menu_Windows = QtWidgets.QMenu(self.menubar)
        self.menu_Windows.setObjectName("menu_Windows")
        self.menu_Help = QtWidgets.QMenu(self.menubar)
        self.menu_Help.setObjectName("menu_Help")
        LinksBrowserWindow.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(LinksBrowserWindow)
        self.statusbar.setObjectName("statusbar")
        LinksBrowserWindow.setStatusBar(self.statusbar)
        self.action_Copy = QtWidgets.QAction(LinksBrowserWindow)
        self.action_Copy.setObjectName("action_Copy")
        self.action_Paste = QtWidgets.QAction(LinksBrowserWindow)
        self.action_Paste.setObjectName("action_Paste")
        self.action_Quit = QtWidgets.QAction(LinksBrowserWindow)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/close"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.action_Quit.setIcon(icon)
        self.action_Quit.setObjectName("action_Quit")
        self.action_Sutta_Search = QtWidgets.QAction(LinksBrowserWindow)
        icon1 = QtGui.QIcon()
        icon1.addPixmap(QtGui.QPixmap(":/book"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.action_Sutta_Search.setIcon(icon1)
        self.action_Sutta_Search.setObjectName("action_Sutta_Search")
        self.action_Dictionary_Search = QtWidgets.QAction(LinksBrowserWindow)
        icon2 = QtGui.QIcon()
        icon2.addPixmap(QtGui.QPixmap(":/dictionary"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.action_Dictionary_Search.setIcon(icon2)
        self.action_Dictionary_Search.setObjectName("action_Dictionary_Search")
        self.action_About = QtWidgets.QAction(LinksBrowserWindow)
        self.action_About.setObjectName("action_About")
        self.action_Website = QtWidgets.QAction(LinksBrowserWindow)
        self.action_Website.setObjectName("action_Website")
        self.action_Close_Window = QtWidgets.QAction(LinksBrowserWindow)
        self.action_Close_Window.setObjectName("action_Close_Window")
        self.action_Open = QtWidgets.QAction(LinksBrowserWindow)
        self.action_Open.setObjectName("action_Open")
        self.action_Document_Reader = QtWidgets.QAction(LinksBrowserWindow)
        self.action_Document_Reader.setObjectName("action_Document_Reader")
        self.action_Library = QtWidgets.QAction(LinksBrowserWindow)
        self.action_Library.setObjectName("action_Library")
        self.action_Memos = QtWidgets.QAction(LinksBrowserWindow)
        self.action_Memos.setObjectName("action_Memos")
        self.action_Dictionaries_Manager = QtWidgets.QAction(LinksBrowserWindow)
        self.action_Dictionaries_Manager.setObjectName("action_Dictionaries_Manager")
        self.action_Links = QtWidgets.QAction(LinksBrowserWindow)
        self.action_Links.setObjectName("action_Links")
        self.menu_File.addAction(self.action_Open)
        self.menu_File.addAction(self.action_Close_Window)
        self.menu_File.addSeparator()
        self.menu_File.addAction(self.action_Quit)
        self.menu_Edit.addAction(self.action_Copy)
        self.menu_Edit.addAction(self.action_Paste)
        self.menu_Windows.addAction(self.action_Sutta_Search)
        self.menu_Windows.addAction(self.action_Dictionary_Search)
        self.menu_Windows.addAction(self.action_Dictionaries_Manager)
        self.menu_Windows.addAction(self.action_Document_Reader)
        self.menu_Windows.addAction(self.action_Library)
        self.menu_Windows.addAction(self.action_Memos)
        self.menu_Windows.addAction(self.action_Links)
        self.menu_Help.addAction(self.action_Website)
        self.menu_Help.addAction(self.action_About)
        self.menubar.addAction(self.menu_File.menuAction())
        self.menubar.addAction(self.menu_Edit.menuAction())
        self.menubar.addAction(self.menu_Windows.menuAction())
        self.menubar.addAction(self.menu_Help.menuAction())

        self.retranslateUi(LinksBrowserWindow)
        QtCore.QMetaObject.connectSlotsByName(LinksBrowserWindow)

    def retranslateUi(self, LinksBrowserWindow):
        _translate = QtCore.QCoreApplication.translate
        LinksBrowserWindow.setWindowTitle(_translate("LinksBrowserWindow", "Links - Simsapa"))
        self.create_link_btn.setText(_translate("LinksBrowserWindow", "Create Link"))
        self.clear_link_btn.setText(_translate("LinksBrowserWindow", "Clear"))
        self.remove_link_btn.setText(_translate("LinksBrowserWindow", "Remove Link"))
        self.set_from_btn.setText(_translate("LinksBrowserWindow", "Set From"))
        self.set_to_btn.setText(_translate("LinksBrowserWindow", "Set To"))
        self.search_button.setText(_translate("LinksBrowserWindow", "Search"))
        self.link_table.setItemText(0, _translate("LinksBrowserWindow", "Suttas"))
        self.link_table.setItemText(1, _translate("LinksBrowserWindow", "DictWords"))
        self.link_table.setItemText(2, _translate("LinksBrowserWindow", "Documents"))
        self.set_page_number.setText(_translate("LinksBrowserWindow", "Set page number:"))
        self.menu_File.setTitle(_translate("LinksBrowserWindow", "&File"))
        self.menu_Edit.setTitle(_translate("LinksBrowserWindow", "&Edit"))
        self.menu_Windows.setTitle(_translate("LinksBrowserWindow", "&Windows"))
        self.menu_Help.setTitle(_translate("LinksBrowserWindow", "&Help"))
        self.action_Copy.setText(_translate("LinksBrowserWindow", "&Copy"))
        self.action_Paste.setText(_translate("LinksBrowserWindow", "&Paste"))
        self.action_Quit.setText(_translate("LinksBrowserWindow", "&Quit"))
        self.action_Quit.setShortcut(_translate("LinksBrowserWindow", "Ctrl+Q"))
        self.action_Sutta_Search.setText(_translate("LinksBrowserWindow", "&Sutta Search"))
        self.action_Sutta_Search.setShortcut(_translate("LinksBrowserWindow", "F5"))
        self.action_Dictionary_Search.setText(_translate("LinksBrowserWindow", "&Dictionary Search"))
        self.action_Dictionary_Search.setShortcut(_translate("LinksBrowserWindow", "F6"))
        self.action_About.setText(_translate("LinksBrowserWindow", "&About"))
        self.action_Website.setText(_translate("LinksBrowserWindow", "&Website"))
        self.action_Close_Window.setText(_translate("LinksBrowserWindow", "&Close Window"))
        self.action_Open.setText(_translate("LinksBrowserWindow", "&Open..."))
        self.action_Open.setShortcut(_translate("LinksBrowserWindow", "Ctrl+O"))
        self.action_Document_Reader.setText(_translate("LinksBrowserWindow", "D&ocument Reader"))
        self.action_Document_Reader.setToolTip(_translate("LinksBrowserWindow", "Document Reader"))
        self.action_Document_Reader.setShortcut(_translate("LinksBrowserWindow", "F7"))
        self.action_Library.setText(_translate("LinksBrowserWindow", "&Library"))
        self.action_Library.setShortcut(_translate("LinksBrowserWindow", "F8"))
        self.action_Memos.setText(_translate("LinksBrowserWindow", "&Memos"))
        self.action_Memos.setToolTip(_translate("LinksBrowserWindow", "Memos"))
        self.action_Memos.setShortcut(_translate("LinksBrowserWindow", "F9"))
        self.action_Dictionaries_Manager.setText(_translate("LinksBrowserWindow", "Dictionaries &Manager"))
        self.action_Dictionaries_Manager.setShortcut(_translate("LinksBrowserWindow", "F10"))
        self.action_Links.setText(_translate("LinksBrowserWindow", "&Links"))
        self.action_Links.setShortcut(_translate("LinksBrowserWindow", "F11"))
from simsapa.assets import icons_rc




PITAKA_GROUPS = {
    'sutta pitaka': {

        'digha nikaya': {
            'silakkhandha vagga': {},
            'maha vagga': {},
            'patika vagga': {},
        },

        'majjhima nikaya': {
            'mula pannasa': [
                'mulapariyaya vagga',
                'sihanada vagga',
                'opamma vagga',
                'mahayamaka vagga',
                'culayamaka vagga',
            ],
            'majjhima pannasa': [
                'gahapati vagga',
                'bhikkhu vagga',
                'paribbajaka vagga',
                'raja vagga',
                'brahmana vagga',
            ],
            'upari pannasa': [
                'devadaha vagga',
                'anupada vagga',
                'sunnata vagga',
                'vibhanga vagga',
                'salayatana vagga',
            ],
        },

        'samyutta nikaya': {
            'sagatha vagga': [
                'sn 1', 'sn 2', 'sn 3', 'sn 4', 'sn 5', 'sn 6', 'sn 7', 'sn 8', 'sn 9', 'sn 10', 'sn 11',
            ],
            'nidana vagga': [
                'sn 12', 'sn 13', 'sn 14', 'sn 15', 'sn 16', 'sn 17', 'sn 18', 'sn 19', 'sn 20', 'sn 21',
            ],
            'khandha vagga': [
                'sn 22', 'sn 23', 'sn 24', 'sn 25', 'sn 26', 'sn 27', 'sn 28', 'sn 29', 'sn 30', 'sn 31', 'sn 32',
                'sn 33', 'sn 34',
            ],
            'salayatana vagga': [
                'sn 35', 'sn 36', 'sn 37', 'sn 38', 'sn 39', 'sn 40', 'sn 41', 'sn 42', 'sn 43', 'sn 44',
            ],
            'maha vagga': [
                'sn 45', 'sn 46', 'sn 47', 'sn 48', 'sn 49', 'sn 50', 'sn 51', 'sn 52', 'sn 53', 'sn 54', 'sn 55',
                'sn 56',
            ],
        },
        'anguttara nikaya': {},
        'khuddaka nikaya': {},
    },

    'vinaya pitaka': {},

    'abhidhamma pitaka': {},
}



import os
import os.path
import logging as _logging
from pathlib import Path
from typing import Optional, Union

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from PyQt5.QtGui import QClipboard

from .db.search import SearchIndexed

from .db import appdata_models as Am
from .db import userdata_models as Um

from simsapa import APP_DB_PATH, USER_DB_PATH, SIMSAPA_DIR, ASSETS_DIR
from simsapa.app.helpers import find_or_create_db

logger = _logging.getLogger(__name__)

USutta = Union[Am.Sutta, Um.Sutta]
UDictWord = Union[Am.DictWord, Um.DictWord]
UDeck = Union[Am.Deck, Um.Deck]
UMemo = Union[Am.Memo, Um.Memo]
UDocument = Union[Am.Document, Um.Document]


class AppData:
    def __init__(self,
                 app_clipboard: Optional[QClipboard] = None,
                 app_db_path: Optional[Path] = None,
                 user_db_path: Optional[Path] = None,
                 api_port: Optional[int] = None,
                 silent_index_if_empty: bool = True):

        self.clipboard: Optional[QClipboard] = app_clipboard

        if app_db_path is None:
            app_db_path = self._find_app_data_or_exit()

        if user_db_path is None:
            user_db_path = self._find_user_data_or_create()

        self.api_url: Optional[str] = None

        if api_port:
            self.api_url = f'http://localhost:{api_port}'

        self.sutta_to_open: Optional[USutta] = None
        self.dict_word_to_open: Optional[UDictWord] = None

        self.db_conn, self.db_session = self._connect_to_db(app_db_path, user_db_path)

        self.search_indexed = SearchIndexed()

        if silent_index_if_empty:
            self.search_indexed.index_all(self.db_session, only_if_empty=True)

    def _connect_to_db(self, app_db_path, user_db_path):
        if not os.path.isfile(app_db_path):
            logger.error(f"Database file doesn't exist: {app_db_path}")
            exit(1)

        if not os.path.isfile(user_db_path):
            logger.error(f"Database file doesn't exist: {user_db_path}")
            exit(1)

        try:
            # Create an in-memory database
            engine = create_engine("sqlite+pysqlite://", echo=False)

            db_conn = engine.connect()

            # Attach appdata and userdata
            db_conn.execute(f"ATTACH DATABASE '{app_db_path}' AS appdata;")
            db_conn.execute(f"ATTACH DATABASE '{user_db_path}' AS userdata;")

            Session = sessionmaker(engine)
            Session.configure(bind=engine)
            db_session = Session()
        except Exception as e:
            logger.error("Can't connect to database.")
            print(e)
            exit(1)

        return (db_conn, db_session)

    def clipboard_setText(self, text):
        if self.clipboard is not None:
            self.clipboard.clear()
            self.clipboard.setText(text)

    def clipboard_getText(self) -> Optional[str]:
        if self.clipboard is not None:
            self.clipboard.clear()
            return self.clipboard.text()
        else:
            return None

    def _find_app_data_or_exit(self):
        if not APP_DB_PATH.exists():
            logger.error("Cannot find appdata.sqlite3")
            exit(1)
        else:
            return APP_DB_PATH

    def _find_user_data_or_create(self) -> Path:
        find_or_create_db(USER_DB_PATH, 'userdata')
        return USER_DB_PATH

def create_app_dirs():
    if not SIMSAPA_DIR.exists():
        SIMSAPA_DIR.mkdir(parents=True, exist_ok=True)

    if not ASSETS_DIR.exists():
        ASSETS_DIR.mkdir(parents=True, exist_ok=True)



import logging as _logging
import json
from typing import Callable, List, Optional
from PyQt5 import QtWidgets

from PyQt5.QtCore import pyqtSignal, QItemSelectionModel
from PyQt5.QtWebEngineWidgets import QWebEngineView
from PyQt5.QtWidgets import (QHBoxLayout, QDialog, QListView, QListWidget, QPushButton, QPlainTextEdit, QFormLayout)

from sqlalchemy.sql import func
from simsapa.app.file_doc import FileDoc

from simsapa.app.types import AppData, UDictWord, USutta
from simsapa.layouts.memos_sidebar import MemoPlainListModel

# from ..app.db import appdata_models as Am
from ..app.db import userdata_models as Um

logger = _logging.getLogger(__name__)


class MemoDialog(QDialog):

    accepted = pyqtSignal(dict) # type: ignore

    def __init__(self, text=''):
        super().__init__()
        self.setWindowTitle("Create Memo")

        self.front = QPlainTextEdit()
        self.front.setFixedSize(300, 50)
        self.front.textChanged.connect(self.unlock)

        self.back = QPlainTextEdit(text)
        self.back.setFixedSize(300, 50)
        self.back.textChanged.connect(self.unlock)

        self.add_btn = QPushButton('Add')
        self.add_btn.setDisabled(True)
        self.add_btn.clicked.connect(self.add_pressed)

        self.close_btn = QPushButton('Close')
        self.close_btn.clicked.connect(self.close_pressed)

        form = QFormLayout(self)

        form.addRow('Front', self.front)
        form.addRow('Back', self.back)

        self.buttons_layout = QHBoxLayout()
        self.buttons_layout.addWidget(self.add_btn)
        self.buttons_layout.addWidget(self.close_btn)

        form.addRow(self.buttons_layout)

    def not_blanks(self) -> bool:
        front = self.front.toPlainText()
        back = self.back.toPlainText()
        return front.strip() != '' and back.strip() != ''

    def unlock(self):
        if self.not_blanks():
            self.add_btn.setEnabled(True)
        else:
            self.add_btn.setDisabled(True)

    def add_pressed(self):
        values = {
            'Front': self.front.toPlainText(),
            'Back': self.back.toPlainText(),
        }
        self.accepted.emit(values)
        self.accept()

    def close_pressed(self):
        self.close()


class HasMemoDialog:
    _app_data: AppData
    front_input: QtWidgets.QPlainTextEdit
    back_input: QtWidgets.QPlainTextEdit
    front: QPlainTextEdit
    back: QPlainTextEdit
    features: List[str] = []
    memos_list: QListView
    _current_sutta: Optional[USutta]
    _current_word: Optional[UDictWord]
    file_doc: Optional[FileDoc]
    db_doc: Optional[Um.Document]
    clear_memo_button: QtWidgets.QPushButton
    remove_memo_button: QtWidgets.QPushButton
    rightside_tabs: QtWidgets.QTabWidget
    memos_tab_idx: int
    content_html: QWebEngineView
    model: MemoPlainListModel
    update_memos_list: Callable

    def init_memo_dialog(self):
        self.memo_dialog_fields = {}

    def set_memo_dialog_fields(self, values):
        self.memo_dialog_fields = values

    def handle_create_memo_for_sutta(self):
        if self._current_sutta is None:
            logger.error("Sutta is not set")
            return

        text = self.content_html.selectedText()

        deck = self._app_data.db_session.query(Um.Deck).first()
        if deck is None:
            logger.error("Can't find the deck")
            return

        self.memo_dialog_fields = {
            'Front': '',
            'Back': '',
        }

        d = MemoDialog(text)
        d.accepted.connect(self.set_memo_dialog_fields)
        d.exec_()

        if self.memo_dialog_fields['Front'] == '' or self.memo_dialog_fields['Back'] == '':
            return

        memo = Um.Memo(
            deck_id=deck.id,
            fields_json=json.dumps(self.memo_dialog_fields),
            created_at=func.now(),
        )

        try:
            self._app_data.db_session.add(memo)
            self._app_data.db_session.commit()

            schema = self._current_sutta.metadata.schema

            if self._current_sutta is not None:

                memo_assoc = Um.MemoAssociation(
                    memo_id=memo.id,
                    associated_table=f'{schema}.suttas',
                    associated_id=self._current_sutta.id,
                )

                self._app_data.db_session.add(memo_assoc)
                self._app_data.db_session.commit()

        except Exception as e:
            logger.error(e)

        if 'memos_sidebar' in self.features:
            # Add to model
            if self.model.memos:
                self.model.memos.append(memo)
            else:
                self.model.memos = [memo]

            index = self.model.index(len(self.model.memos) - 1)
            self.memos_list.selectionModel().select(index, QItemSelectionModel.SelectionFlag.Select)

            self.update_memos_list()

    def handle_create_memo_for_dict_word(self):
        if self._current_word is None:
            logger.error("Word is not set")
            return

        text = self.content_html.selectedText()

        deck = self._app_data.db_session.query(Um.Deck).first()
        if deck is None:
            logger.error("Can't find the deck")
            return

        self.memo_dialog_fields = {
            'Front': '',
            'Back': '',
        }

        d = MemoDialog(text)
        d.accepted.connect(self.set_memo_dialog_fields)
        d.exec_()

        if self.memo_dialog_fields['Front'] == '' or self.memo_dialog_fields['Back'] == '':
            return

        memo = Um.Memo(
            deck_id=deck.id,
            fields_json=json.dumps(self.memo_dialog_fields),
            created_at=func.now(),
        )

        try:
            self._app_data.db_session.add(memo)
            self._app_data.db_session.commit()

            schema = self._current_word.metadata.schema

            if self._current_word is not None:

                memo_assoc = Um.MemoAssociation(
                    memo_id=memo.id,
                    associated_table=f'{schema}.dict_words',
                    associated_id=self._current_word.id,
                )

                self._app_data.db_session.add(memo_assoc)
                self._app_data.db_session.commit()

        except Exception as e:
            logger.error(e)

        if 'memos_sidebar' in self.features:
            # Add to model
            if self.model.memos:
                self.model.memos.append(memo)
            else:
                self.model.memos = [memo]

            index = self.model.index(len(self.model.memos) - 1)
            self.memos_list.selectionModel().select(index, QItemSelectionModel.SelectionFlag.Select)

            self.update_memos_list()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'simsapa/assets/ui/import_stardict_dialog.ui'
#
# Created by: PyQt5 UI code generator 5.15.6
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_ImportStarDictDialog(object):
    def setupUi(self, ImportStarDictDialog):
        ImportStarDictDialog.setObjectName("ImportStarDictDialog")
        ImportStarDictDialog.resize(500, 430)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.MinimumExpanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(ImportStarDictDialog.sizePolicy().hasHeightForWidth())
        ImportStarDictDialog.setSizePolicy(sizePolicy)
        ImportStarDictDialog.setBaseSize(QtCore.QSize(500, 400))
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout(ImportStarDictDialog)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.open_btn = QtWidgets.QPushButton(ImportStarDictDialog)
        self.open_btn.setObjectName("open_btn")
        self.verticalLayout.addWidget(self.open_btn, 0, QtCore.Qt.AlignTop)
        self.info = QtWidgets.QLabel(ImportStarDictDialog)
        self.info.setObjectName("info")
        self.verticalLayout.addWidget(self.info)
        spacerItem = QtWidgets.QSpacerItem(20, 0, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem)
        self.tabWidget = QtWidgets.QTabWidget(ImportStarDictDialog)
        self.tabWidget.setObjectName("tabWidget")
        self.create_new_tab = QtWidgets.QWidget()
        self.create_new_tab.setObjectName("create_new_tab")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.create_new_tab)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout()
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.label_3 = QtWidgets.QLabel(self.create_new_tab)
        self.label_3.setObjectName("label_3")
        self.verticalLayout_2.addWidget(self.label_3)
        self.title = QtWidgets.QLineEdit(self.create_new_tab)
        self.title.setMinimumSize(QtCore.QSize(0, 40))
        self.title.setObjectName("title")
        self.verticalLayout_2.addWidget(self.title)
        self.label_4 = QtWidgets.QLabel(self.create_new_tab)
        self.label_4.setObjectName("label_4")
        self.verticalLayout_2.addWidget(self.label_4)
        self.label = QtWidgets.QLineEdit(self.create_new_tab)
        self.label.setMinimumSize(QtCore.QSize(0, 40))
        self.label.setObjectName("label")
        self.verticalLayout_2.addWidget(self.label)
        self.label_5 = QtWidgets.QLabel(self.create_new_tab)
        self.label_5.setObjectName("label_5")
        self.verticalLayout_2.addWidget(self.label_5)
        self.import_new_btn = QtWidgets.QPushButton(self.create_new_tab)
        self.import_new_btn.setEnabled(False)
        self.import_new_btn.setCheckable(False)
        self.import_new_btn.setObjectName("import_new_btn")
        self.verticalLayout_2.addWidget(self.import_new_btn)
        spacerItem1 = QtWidgets.QSpacerItem(20, 0, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_2.addItem(spacerItem1)
        self.verticalLayout_3.addLayout(self.verticalLayout_2)
        self.tabWidget.addTab(self.create_new_tab, "")
        self.update_existing_tab = QtWidgets.QWidget()
        self.update_existing_tab.setObjectName("update_existing_tab")
        self.verticalLayout_5 = QtWidgets.QVBoxLayout(self.update_existing_tab)
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout()
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.label_6 = QtWidgets.QLabel(self.update_existing_tab)
        self.label_6.setObjectName("label_6")
        self.verticalLayout_4.addWidget(self.label_6)
        self.select_title = QtWidgets.QComboBox(self.update_existing_tab)
        self.select_title.setObjectName("select_title")
        self.verticalLayout_4.addWidget(self.select_title)
        self.selected_info = QtWidgets.QLabel(self.update_existing_tab)
        self.selected_info.setObjectName("selected_info")
        self.verticalLayout_4.addWidget(self.selected_info)
        spacerItem2 = QtWidgets.QSpacerItem(20, 0, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_4.addItem(spacerItem2)
        self.update_existing_btn = QtWidgets.QPushButton(self.update_existing_tab)
        self.update_existing_btn.setEnabled(False)
        self.update_existing_btn.setObjectName("update_existing_btn")
        self.verticalLayout_4.addWidget(self.update_existing_btn)
        self.verticalLayout_5.addLayout(self.verticalLayout_4)
        self.tabWidget.addTab(self.update_existing_tab, "")
        self.verticalLayout.addWidget(self.tabWidget, 0, QtCore.Qt.AlignBottom)
        self.msg = QtWidgets.QLabel(ImportStarDictDialog)
        self.msg.setObjectName("msg")
        self.verticalLayout.addWidget(self.msg)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.close_btn = QtWidgets.QPushButton(ImportStarDictDialog)
        self.close_btn.setObjectName("close_btn")
        self.horizontalLayout.addWidget(self.close_btn)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.horizontalLayout_2.addLayout(self.verticalLayout)

        self.retranslateUi(ImportStarDictDialog)
        self.tabWidget.setCurrentIndex(0)
        QtCore.QMetaObject.connectSlotsByName(ImportStarDictDialog)

    def retranslateUi(self, ImportStarDictDialog):
        _translate = QtCore.QCoreApplication.translate
        ImportStarDictDialog.setWindowTitle(_translate("ImportStarDictDialog", "Import from a StarDict Zip Dictionary"))
        self.open_btn.setText(_translate("ImportStarDictDialog", "Select StarDict ZIP..."))
        self.info.setText(_translate("ImportStarDictDialog", "info"))
        self.label_3.setText(_translate("ImportStarDictDialog", "Title:"))
        self.label_4.setText(_translate("ImportStarDictDialog", "Unique short label:"))
        self.label_5.setText(_translate("ImportStarDictDialog", "<html><head/><body><p>The label (such as NCPED) is used to distinguish the same words</p><p>in different dictionaries.</p></body></html>"))
        self.import_new_btn.setText(_translate("ImportStarDictDialog", "Import as new"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.create_new_tab), _translate("ImportStarDictDialog", "Import as new"))
        self.label_6.setText(_translate("ImportStarDictDialog", "Select title:"))
        self.selected_info.setText(_translate("ImportStarDictDialog", "selected info"))
        self.update_existing_btn.setText(_translate("ImportStarDictDialog", "Update existing"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.update_existing_tab), _translate("ImportStarDictDialog", "Update existing"))
        self.msg.setText(_translate("ImportStarDictDialog", "status msg"))
        self.close_btn.setText(_translate("ImportStarDictDialog", "Close"))



import sys
from enum import Enum
import typer

app = typer.Typer()
index_app = typer.Typer()
app.add_typer(index_app, name="index")

@app.command()
def gui():
    from simsapa.gui import start
    start()

class QueryType(str, Enum):
    suttas = "suttas"
    dict_words = "dict_words"

@app.command()
def query(query_type: QueryType, query: str, print_titles: bool = True, print_count: bool = False):
    """Query the database."""

    from simsapa.app.db import search
    from simsapa.app.types import AppData

    app_data = AppData()

    if query_type == QueryType.suttas:
        search_query = search.SearchQuery(
            app_data.search_indexed.suttas_index,
            20,
            search.sutta_hit_to_search_result,
        )
    elif query_type == QueryType.dict_words:
        search_query = search.SearchQuery(
            app_data.search_indexed.dict_words_index,
            20,
            search.dict_word_hit_to_search_result,
        )
    else:
        print("Unrecognized query type.")
        return

    search_query.new_query(query)

    if print_count:
        print(f"Results count: {search_query.hits}")

    if print_titles:
        for i in search_query.all_results(highlight=False):
            print(i['title'])

@index_app.command("create")
def index_create():
    """Create database indexes, removing existing ones."""
    from simsapa.app.db.search import SearchIndexed
    search_indexed = SearchIndexed()
    search_indexed.create_all()

@index_app.command("reindex")
def index_reindex():
    """Clear and rebuild database indexes."""
    from simsapa.app.db.search import SearchIndexed
    from simsapa.app.types import AppData
    search_indexed = SearchIndexed()
    app_data = AppData()

    search_indexed.create_all()
    search_indexed.index_all(app_data.db_session)

def main():
    if len(sys.argv) == 1:
        gui()
    else:
        app()

if __name__ == "__main__":
    main()



from functools import partial
import math
from typing import List, Optional
from markdown import markdown
# from sqlalchemy.orm import joinedload
from pathlib import Path
import json
import queue

from PyQt5.QtCore import Qt, QUrl, QTimer
from PyQt5.QtGui import QKeySequence, QCloseEvent
from PyQt5.QtWidgets import (QLabel, QListWidget, QMainWindow, QAction,
                             QVBoxLayout, QHBoxLayout, QPushButton,
                             QSizePolicy)
from PyQt5.QtWebEngineWidgets import QWebEngineView

from simsapa import ASSETS_DIR, APP_QUEUES
from ..app.db import appdata_models as Am
from ..app.db import userdata_models as Um
from ..app.db.search import SearchQuery, SearchResult, dict_word_hit_to_search_result
from ..app.types import AppData, USutta, UDictWord
from ..assets.ui.dictionary_search_window_ui import Ui_DictionarySearchWindow
from .memo_dialog import HasMemoDialog
from .memos_sidebar import HasMemosSidebar
from .links_sidebar import HasLinksSidebar
from .results_list import HasResultsList
from .html_content import html_page
from .import_stardict_dialog import HasImportStarDictDialog


class DictionarySearchWindow(QMainWindow, Ui_DictionarySearchWindow,
                             HasMemoDialog, HasLinksSidebar, HasMemosSidebar,
                             HasResultsList, HasImportStarDictDialog):

    def __init__(self, app_data: AppData, parent=None) -> None:
        super().__init__(parent)
        self.setupUi(self)

        self.results_list: QListWidget
        self.recent_list: QListWidget

        self.features: List[str] = []
        self._app_data: AppData = app_data
        self._results: List[SearchResult] = []
        self._recent: List[UDictWord] = []

        self._current_word: Optional[UDictWord] = None

        self.page_len = 20
        self.search_query = SearchQuery(
            self._app_data.search_indexed.dict_words_index,
            self.page_len,
            dict_word_hit_to_search_result,
        )

        self.queue_id = 'window_' + str(len(APP_QUEUES))
        APP_QUEUES[self.queue_id] = queue.Queue()
        self.messages_url = f'{self._app_data.api_url}/queues/{self.queue_id}'

        self.graph_path: Path = ASSETS_DIR.joinpath(f"{self.queue_id}.html")

        self.timer = QTimer()
        self.timer.timeout.connect(self.handle_messages)
        self.timer.start(300)

        self._ui_setup()
        self._connect_signals()

        self.init_results_list()
        self.init_memo_dialog()
        self.init_memos_sidebar()
        self.init_links_sidebar()
        self.init_stardict_import_dialog()

        self._setup_content_html_context_menu()

        self.statusbar.showMessage("Ready", 3000)

    def closeEvent(self, event: QCloseEvent):
        if self.queue_id in APP_QUEUES.keys():
            del APP_QUEUES[self.queue_id]

        if self.graph_path.exists():
            self.graph_path.unlink()

        event.accept()

    def handle_messages(self):
        if self.queue_id in APP_QUEUES.keys():
            try:
                s = APP_QUEUES[self.queue_id].get_nowait()
                data = json.loads(s)
                if data['action'] == 'show_sutta':
                    self._show_sutta_from_message(data['arg'])

                elif data['action'] == 'show_sutta_by_uid':
                    info = data['arg']
                    if 'uid' in info.keys():
                        self._show_sutta_by_uid(info['uid'])

                elif data['action'] == 'show_word_by_uid':
                    info = data['arg']
                    if 'uid' in info.keys():
                        self._show_word_by_uid(info['uid'])

                APP_QUEUES[self.queue_id].task_done()
            except queue.Empty:
                pass

    def _ui_setup(self):
        self.status_msg = QLabel("Word title")
        self.statusbar.addPermanentWidget(self.status_msg)

        self.links_tab_idx = 1
        self.memos_tab_idx = 2

        self._setup_pali_buttons()
        self._setup_content_html()

        self.search_input.setFocus()

    def _setup_content_html(self):
        self.content_html = QWebEngineView()
        self.content_html.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.content_html.setHtml('')
        self.content_html.show()
        self.content_layout.addWidget(self.content_html)

    def _setup_pali_buttons(self):
        self.pali_buttons_layout = QVBoxLayout()
        self.searchbar_layout.addLayout(self.pali_buttons_layout)

        lowercase = 'ā ī ū ṃ ṁ ṅ ñ ṭ ḍ ṇ ḷ'.split(' ')
        uppercase = "Ā Ī Ū Ṃ Ṁ Ṅ Ñ Ṭ Ḍ Ṇ Ḷ".split(' ')

        lowercase_row = QHBoxLayout()

        for i in lowercase:
            btn = QPushButton(i)
            btn.setFixedSize(30, 30)
            btn.clicked.connect(partial(self._append_to_query, i))
            lowercase_row.addWidget(btn)

        uppercase_row = QHBoxLayout()

        for i in uppercase:
            btn = QPushButton(i)
            btn.setFixedSize(30, 30)
            btn.clicked.connect(partial(self._append_to_query, i))
            uppercase_row.addWidget(btn)

        self.pali_buttons_layout.addLayout(lowercase_row)
        self.pali_buttons_layout.addLayout(uppercase_row)

    def _append_to_query(self, s: str):
        a = self.search_input.text()
        self.search_input.setText(a + s)
        self.search_input.setFocus()

    def _handle_query(self, min_length=4):
        self._highlight_query()
        query = self.search_input.text()

        if len(query) >= min_length:
            self._results = self._word_search_query(query)

            if self.search_query.hits > 0:
                self.rightside_tabs.setTabText(0, f"Results ({self.search_query.hits})")
            else:
                self.rightside_tabs.setTabText(0, "Results")

            self.render_results_page()

    def _set_content_html(self, html):
        self.content_html.setHtml(html)
        self._highlight_query()

    def _highlight_query(self):
        query = self.search_input.text()
        self.content_html.findText(query)

    def _add_recent(self, word: UDictWord):
        # de-duplicate: if item already exists, remove it
        if word in self._recent:
            self._recent.remove(word)
        # insert new item on top
        self._recent.insert(0, word)

        # Rebuild Qt recents list
        self.recent_list.clear()
        words = list(map(lambda x: x.word, self._recent))
        self.recent_list.insertItems(0, words) # type: ignore

    def _dict_word_from_result(self, x: SearchResult) -> Optional[UDictWord]:
        if x['schema_name'] == 'appdata':
            word = self._app_data.db_session \
                                 .query(Am.DictWord) \
                                 .filter(Am.DictWord.id == x['db_id']) \
                                 .first()
        else:
            word = self._app_data.db_session \
                                 .query(Um.DictWord) \
                                 .filter(Um.DictWord.id == x['db_id']) \
                                 .first()
        return word

    def _handle_result_select(self):
        selected_idx = self.results_list.currentRow()
        if selected_idx < len(self._results):
            word = self._dict_word_from_result(self._results[selected_idx])
            if word is not None:
                self._show_word(word)
                self._add_recent(word)

    def _handle_recent_select(self):
        selected_idx = self.recent_list.currentRow()
        word: UDictWord = self._recent[selected_idx]
        self._show_word(word)

    def _show_word(self, word: UDictWord):
        self._current_word = word
        self.status_msg.setText(word.word) # type: ignore

        self.update_memos_list_for_dict_word(word)
        self.show_network_graph(word)

        def example_format(example):
            return "<div>" + example.text_html + "</div><div>" + example.translation_html + "</div>"

        examples = "".join(list(map(example_format, word.examples))) # type: ignore

        if word.definition_html is not None and word.definition_html != '':
            definition = word.definition_html
        elif word.definition_plain is not None and word.definition_plain != '':
            definition = markdown(word.definition_plain) # type: ignore
        else:
            definition = '<p>No definition.</p>'

        messages_url = f'{self._app_data.api_url}/queues/{self.queue_id}'
        content = "<div>%s</div><div>%s</div>" % (definition, examples)
        html = html_page(content, messages_url)

        # show the word content
        self._set_content_html(html)

    def show_network_graph(self, word: UDictWord):
        self.generate_graph_for_dict_word(word, self.queue_id, self.graph_path, self.messages_url)
        self.content_graph.load(QUrl('file://' + str(self.graph_path.absolute())))

    def _word_search_query(self, query: str) -> List[SearchResult]:
        results = self.search_query.new_query(query)
        hits = self.search_query.hits

        if hits == 0:
            self.results_page_input.setMinimum(0)
            self.results_page_input.setMaximum(0)
            self.results_first_page_btn.setEnabled(False)
            self.results_last_page_btn.setEnabled(False)

        elif hits <= self.page_len:
            self.results_page_input.setMinimum(1)
            self.results_page_input.setMaximum(1)
            self.results_first_page_btn.setEnabled(False)
            self.results_last_page_btn.setEnabled(False)

        else:
            pages = math.floor(hits / self.page_len) + 1
            self.results_page_input.setMinimum(1)
            self.results_page_input.setMaximum(pages)
            self.results_first_page_btn.setEnabled(True)
            self.results_last_page_btn.setEnabled(True)

        return results

    def _show_sutta_from_message(self, info):
        sutta: Optional[USutta] = None

        if info['table'] == 'appdata.suttas':
            sutta = self._app_data.db_session \
                .query(Am.Sutta) \
                .filter(Am.Sutta.id == info['id']) \
                .first()

        elif info['table'] == 'userdata.suttas':
            sutta = self._app_data.db_session \
                .query(Um.Sutta) \
                .filter(Um.Sutta.id == info['id']) \
                .first()

        self._app_data.sutta_to_open = sutta
        self.action_Sutta_Search.activate(QAction.Trigger) # type: ignore

    def _show_sutta_by_uid(self, uid: str):
        results: List[USutta] = []

        res = self._app_data.db_session \
            .query(Am.Sutta) \
            .filter(Am.Sutta.uid == uid) \
            .all()
        results.extend(res)

        res = self._app_data.db_session \
            .query(Um.Sutta) \
            .filter(Um.Sutta.uid == uid) \
            .all()
        results.extend(res)

        if len(results) > 0:
            self._app_data.sutta_to_open = results[0]
            self.action_Sutta_Search.activate(QAction.Trigger) # type: ignore

    def _show_word_by_uid(self, uid: str):
        results: List[UDictWord] = []

        res = self._app_data.db_session \
            .query(Am.DictWord) \
            .filter(Am.DictWord.uid == uid) \
            .all()
        results.extend(res)

        res = self._app_data.db_session \
            .query(Um.DictWord) \
            .filter(Um.DictWord.uid == uid) \
            .all()
        results.extend(res)

        if len(results) > 0:
            self._show_word(results[0])

    def _handle_copy(self):
        text = self.content_html.selectedText()
        # U+2029 Paragraph Separator to blank line
        text = text.replace('\u2029', "\n\n")
        self._app_data.clipboard_setText(text)

    def _setup_content_html_context_menu(self):
        self.content_html.setContextMenuPolicy(Qt.ActionsContextMenu) # type: ignore

        copyAction = QAction("Copy", self.content_html)
        copyAction.setShortcut(QKeySequence("Ctrl+C"))
        copyAction.triggered.connect(partial(self._handle_copy))

        self.content_html.addAction(copyAction)

        memoAction = QAction("Create Memo", self.content_html)
        memoAction.setShortcut(QKeySequence("Ctrl+M"))
        memoAction.triggered.connect(partial(self.handle_create_memo_for_dict_word))

        self.content_html.addAction(memoAction)

    def _connect_signals(self):
        self.action_Close_Window \
            .triggered.connect(partial(self.close))

        self.search_button.clicked.connect(partial(self._handle_query, min_length=1))
        self.search_input.textChanged.connect(partial(self._handle_query, min_length=4))
        # self.search_input.returnPressed.connect(partial(self._update_result))

        self.recent_list.itemSelectionChanged.connect(partial(self._handle_recent_select))

        self.add_memo_button \
            .clicked.connect(partial(self.add_memo_for_dict_word))

        self.action_Import_from_StarDict \
            .triggered.connect(partial(self.show_import_from_stardict_dialog))



import sys
import os
import traceback
import logging as _logging
import logging.config
import yaml
import threading

from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import (QApplication, QSystemTrayIcon, QMenu, QAction)

from simsapa import APP_DB_PATH
from .app.types import AppData, create_app_dirs
from .app.windows import AppWindows
from .app.api import start_server, find_available_port
from .layouts.download_appdata import DownloadAppdataWindow
from .layouts.error_message import ErrorMessageWindow

from simsapa.assets import icons_rc  # noqa: F401

logger = _logging.getLogger(__name__)

if os.path.exists("logging.yaml"):
    with open("logging.yaml", 'r') as f:
        config = yaml.safe_load(f.read())
        _logging.config.dictConfig(config) # type: ignore


def excepthook(exc_type, exc_value, exc_tb):
    tb = "".join(traceback.format_exception(exc_type, exc_value, exc_tb))
    logger.error("Error:\n", tb)
    w = ErrorMessageWindow(user_message=None, debug_info=tb)
    w.show()


sys.excepthook = excepthook


def start():
    logger.info("start()")

    create_app_dirs()

    if not APP_DB_PATH.exists():
        dl_app = QApplication(sys.argv)
        w = DownloadAppdataWindow()
        w.show()
        status = dl_app.exec_()
        logger.info(f"main() Exiting with status {status}.")
        sys.exit(status)

    app = QApplication(sys.argv)

    port = find_available_port()
    daemon = threading.Thread(name='daemon_server',
                              target=start_server,
                              args=(port,))
    daemon.setDaemon(True)
    daemon.start()

    app_data = AppData(app_clipboard=app.clipboard(), api_port=port)

    # === Create systray ===

    app.setQuitOnLastWindowClosed(False)

    tray = QSystemTrayIcon(QIcon(":simsapa-tray"))
    tray.setVisible(True)

    menu = QMenu()

    ac = QAction(QIcon(":close"), "Quit")
    ac.triggered.connect(app.quit)
    menu.addAction(ac)

    tray.setContextMenu(menu)

    # === Create first window ===

    app_windows = AppWindows(app, app_data)
    app_windows._new_sutta_search_window()

    status = app.exec_()
    logger.info(f"main() Exiting with status {status}.")
    sys.exit(status)



import logging as _logging
import json
import math
from pathlib import Path
import queue

from functools import partial
from typing import List, Optional

from PyQt5.QtCore import Qt, QUrl, QTimer
from PyQt5.QtGui import QKeySequence, QCloseEvent
from PyQt5.QtWidgets import (QLabel, QMainWindow, QAction, QListWidgetItem,
                             QVBoxLayout, QHBoxLayout, QPushButton,
                             QSizePolicy, QListWidget)
from PyQt5.QtWebEngineWidgets import QWebEngineView
from whoosh.searching import Hit

from simsapa import ASSETS_DIR, APP_QUEUES
from ..app.db.search import SearchResult, SearchQuery, sutta_hit_to_search_result
from ..app.db import appdata_models as Am
from ..app.db import userdata_models as Um
from ..app.types import AppData, USutta, UDictWord
from ..assets.ui.sutta_search_window_ui import Ui_SuttaSearchWindow
from .search_item import SearchItemWidget
from .memo_dialog import HasMemoDialog
from .memos_sidebar import HasMemosSidebar
from .links_sidebar import HasLinksSidebar
from .results_list import HasResultsList
from .html_content import html_page

logger = _logging.getLogger(__name__)


class SuttaSearchWindow(QMainWindow, Ui_SuttaSearchWindow, HasMemoDialog,
                        HasLinksSidebar, HasMemosSidebar, HasResultsList):

    def __init__(self, app_data: AppData, parent=None) -> None:
        super().__init__(parent)
        self.setupUi(self)

        self.results_list: QListWidget
        self.recent_list: QListWidget

        self.features: List[str] = []
        self._app_data: AppData = app_data
        self._results: List[SearchResult] = []
        self._recent: List[USutta] = []

        self._current_sutta: Optional[USutta] = None

        self.page_len = 20
        self.search_query = SearchQuery(
            self._app_data.search_indexed.suttas_index,
            self.page_len,
            sutta_hit_to_search_result,
        )

        self.queue_id = 'window_' + str(len(APP_QUEUES))
        APP_QUEUES[self.queue_id] = queue.Queue()
        self.messages_url = f'{self._app_data.api_url}/queues/{self.queue_id}'

        self.graph_path: Path = ASSETS_DIR.joinpath(f"{self.queue_id}.html")

        self.timer = QTimer()
        self.timer.timeout.connect(self.handle_messages)
        self.timer.start(300)

        self._ui_setup()
        self._connect_signals()

        self.init_results_list()
        self.init_memo_dialog()
        self.init_memos_sidebar()
        self.init_links_sidebar()

        self._setup_content_html_context_menu()

        self.statusbar.showMessage("Ready", 3000)

    def closeEvent(self, event: QCloseEvent):
        if self.queue_id in APP_QUEUES.keys():
            del APP_QUEUES[self.queue_id]

        if self.graph_path.exists():
            self.graph_path.unlink()

        event.accept()

    def handle_messages(self):
        if self.queue_id in APP_QUEUES.keys():
            try:
                s = APP_QUEUES[self.queue_id].get_nowait()
                data = json.loads(s)
                if data['action'] == 'show_sutta':
                    self._show_sutta_from_message(data['arg'])

                elif data['action'] == 'show_sutta_by_uid':
                    info = data['arg']
                    if 'uid' in info.keys():
                        self._show_sutta_by_uid(info['uid'])

                elif data['action'] == 'show_word_by_uid':
                    info = data['arg']
                    if 'uid' in info.keys():
                        self._show_word_by_uid(info['uid'])

                APP_QUEUES[self.queue_id].task_done()
            except queue.Empty:
                pass

    def _ui_setup(self):
        self.status_msg = QLabel("Sutta title")
        self.statusbar.addPermanentWidget(self.status_msg)

        self.links_tab_idx = 1
        self.memos_tab_idx = 2

        self._setup_pali_buttons()
        self._setup_content_html()

        self.search_input.setFocus()

    def _setup_content_html(self):
        self.content_html = QWebEngineView()
        self.content_html.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.content_html.setHtml('')
        self.content_html.show()
        self.content_layout.addWidget(self.content_html)

    def _setup_pali_buttons(self):
        self.pali_buttons_layout = QVBoxLayout()
        self.searchbar_layout.addLayout(self.pali_buttons_layout)

        lowercase = 'ā ī ū ṃ ṁ ṅ ñ ṭ ḍ ṇ ḷ'.split(' ')
        uppercase = "Ā Ī Ū Ṃ Ṁ Ṅ Ñ Ṭ Ḍ Ṇ Ḷ".split(' ')

        lowercase_row = QHBoxLayout()

        for i in lowercase:
            btn = QPushButton(i)
            btn.setFixedSize(30, 30)
            btn.clicked.connect(partial(self._append_to_query, i))
            lowercase_row.addWidget(btn)

        uppercase_row = QHBoxLayout()

        for i in uppercase:
            btn = QPushButton(i)
            btn.setFixedSize(30, 30)
            btn.clicked.connect(partial(self._append_to_query, i))
            uppercase_row.addWidget(btn)

        self.pali_buttons_layout.addLayout(lowercase_row)
        self.pali_buttons_layout.addLayout(uppercase_row)

    def _append_to_query(self, s: str):
        a = self.search_input.text()
        self.search_input.setText(a + s)
        self.search_input.setFocus()

    def _handle_query(self, min_length=4):
        self._highlight_query()
        query = self.search_input.text()

        if len(query) >= min_length:
            self._results = self._sutta_search_query(query)

            if self.search_query.hits > 0:
                self.rightside_tabs.setTabText(0, f"Results ({self.search_query.hits})")
            else:
                self.rightside_tabs.setTabText(0, "Results")

            self.render_results_page()

    def _set_content_html(self, html):
        self.content_html.setHtml(html)
        self._highlight_query()

    def _highlight_query(self):
        query = self.search_input.text()
        self.content_html.findText(query)

    def _add_recent(self, sutta: USutta):
        # de-duplicate: if item already exists, remove it
        if sutta in self._recent:
            self._recent.remove(sutta)
        # insert new item on top
        self._recent.insert(0, sutta)

        # Rebuild Qt recents list
        self.recent_list.clear()
        titles = list(map(lambda x: x.title, self._recent))
        self.recent_list.insertItems(0, titles) # type: ignore

    def _sutta_from_result(self, x: SearchResult) -> Optional[USutta]:
        if x['schema_name'] == 'appdata':
            sutta = self._app_data.db_session \
                                  .query(Am.Sutta) \
                                  .filter(Am.Sutta.id == x['db_id']) \
                                  .first()
        else:
            sutta = self._app_data.db_session \
                                  .query(Um.Sutta) \
                                  .filter(Um.Sutta.id == x['db_id']) \
                                  .first()
        return sutta

    def _handle_result_select(self):
        selected_idx = self.results_list.currentRow()
        if selected_idx < len(self._results):
            sutta = self._sutta_from_result(self._results[selected_idx])
            if sutta is not None:
                self._show_sutta(sutta)
                self._add_recent(sutta)

    def _handle_recent_select(self):
        selected_idx = self.recent_list.currentRow()
        sutta: USutta = self._recent[selected_idx]
        self._show_sutta(sutta)

    def _show_sutta_from_message(self, info):
        sutta: Optional[USutta] = None

        if info['table'] == 'appdata.suttas':
            sutta = self._app_data.db_session \
                .query(Am.Sutta) \
                .filter(Am.Sutta.id == info['id']) \
                .first()

        elif info['table'] == 'userdata.suttas':
            sutta = self._app_data.db_session \
                .query(Um.Sutta) \
                .filter(Um.Sutta.id == info['id']) \
                .first()

        if sutta:
            self._show_sutta(sutta)

    def _show_sutta_by_uid(self, uid: str):
        results: List[USutta] = []

        res = self._app_data.db_session \
            .query(Am.Sutta) \
            .filter(Am.Sutta.uid == uid) \
            .all()
        results.extend(res)

        res = self._app_data.db_session \
            .query(Um.Sutta) \
            .filter(Um.Sutta.uid == uid) \
            .all()
        results.extend(res)

        if len(results) > 0:
            self._show_sutta(results[0])

    def _show_word_by_uid(self, uid: str):
        results: List[UDictWord] = []

        res = self._app_data.db_session \
            .query(Am.DictWord) \
            .filter(Am.DictWord.uid == uid) \
            .all()
        results.extend(res)

        res = self._app_data.db_session \
            .query(Um.DictWord) \
            .filter(Um.DictWord.uid == uid) \
            .all()
        results.extend(res)

        if len(results) > 0:
            self._app_data.dict_word_to_open = results[0]
            self.action_Dictionary_Search.activate(QAction.ActionEvent.Trigger)

    def _show_sutta(self, sutta: USutta):
        self._current_sutta = sutta
        self.status_msg.setText(sutta.title) # type: ignore

        self.update_memos_list_for_sutta(sutta)
        self.show_network_graph(sutta)

        if sutta.content_html is not None and sutta.content_html != '':
            content = sutta.content_html
        elif sutta.content_plain is not None and sutta.content_plain != '':
            content = '<pre>' + sutta.content_plain + '</pre>'
        else:
            content = 'No content.'

        html = html_page(content, self.messages_url) # type: ignore

        # show the sutta content
        self._set_content_html(html)

    def show_network_graph(self, sutta: USutta):
        self.generate_graph_for_sutta(sutta, self.queue_id, self.graph_path, self.messages_url)
        self.content_graph.load(QUrl('file://' + str(self.graph_path.absolute())))

    def _sutta_search_query(self, query: str) -> List[SearchResult]:
        results = self.search_query.new_query(query)
        hits = self.search_query.hits

        if hits == 0:
            self.results_page_input.setMinimum(0)
            self.results_page_input.setMaximum(0)
            self.results_first_page_btn.setEnabled(False)
            self.results_last_page_btn.setEnabled(False)

        elif hits <= self.page_len:
            self.results_page_input.setMinimum(1)
            self.results_page_input.setMaximum(1)
            self.results_first_page_btn.setEnabled(False)
            self.results_last_page_btn.setEnabled(False)

        else:
            pages = math.floor(hits / self.page_len) + 1
            self.results_page_input.setMinimum(1)
            self.results_page_input.setMaximum(pages)
            self.results_first_page_btn.setEnabled(True)
            self.results_last_page_btn.setEnabled(True)

        return results

    def _handle_copy(self):
        text = self.content_html.selectedText()
        # U+2029 Paragraph Separator to blank line
        text = text.replace('\u2029', "\n\n")
        self._app_data.clipboard_setText(text)

    def _setup_content_html_context_menu(self):
        self.content_html.setContextMenuPolicy(Qt.ContextMenuPolicy.ActionsContextMenu)

        copyAction = QAction("Copy", self.content_html)
        copyAction.setShortcut(QKeySequence("Ctrl+C"))
        copyAction.triggered.connect(partial(self._handle_copy))

        self.content_html.addAction(copyAction)

        memoAction = QAction("Create Memo", self.content_html)
        memoAction.setShortcut(QKeySequence("Ctrl+M"))
        memoAction.triggered.connect(partial(self.handle_create_memo_for_sutta))

        self.content_html.addAction(memoAction)

    def _connect_signals(self):
        self.action_Close_Window \
            .triggered.connect(partial(self.close))

        self.search_button.clicked.connect(partial(self._handle_query, min_length=1))
        self.search_input.textChanged.connect(partial(self._handle_query, min_length=4))
        # self.search_input.returnPressed.connect(partial(self._update_result))

        self.recent_list.itemSelectionChanged.connect(partial(self._handle_recent_select))

        self.add_memo_button \
            .clicked.connect(partial(self.add_memo_for_sutta))



import logging as _logging

from functools import partial
from typing import Callable, List
from PyQt5.QtGui import QColor

from PyQt5.QtWidgets import QLabel, QListWidget, QListWidgetItem, QPushButton, QSpinBox

from simsapa.app.db.search import SearchResult, SearchQuery
from simsapa.layouts.search_item import SearchItemWidget

logger = _logging.getLogger(__name__)

class HasResultsList:
    features: List[str]
    results_label: QLabel
    results_list: QListWidget
    results_last_page_btn: QPushButton
    results_first_page_btn: QPushButton
    results_page_input: QSpinBox
    search_query: SearchQuery
    _results: List[SearchResult]
    _handle_query: Callable
    _handle_result_select: Callable
    page_len: int

    def init_results_list(self):
        self.features.append('results_list')
        self.connect_results_list_signals()

        self.results_label.clear()
        self.results_list.clear()

        self.results_page_input.setMinimum(0)
        self.results_page_input.setMaximum(0)
        self.results_first_page_btn.setEnabled(False)
        self.results_last_page_btn.setEnabled(False)

    def render_results_page(self):
        page_num = self.results_page_input.value() - 1
        page_start = page_num * self.page_len

        page_end = page_start + self.page_len
        if page_end > self.search_query.hits:
            page_end = self.search_query.hits

        self.results_list.clear()

        if self.search_query.hits == 0:
            self.results_label.clear()
            return

        self._results = self.search_query.highlight_results_page(page_num)

        msg = f"Showing {page_start+1}-{page_end} out of {self.search_query.hits}"
        self.results_label.setText(msg)

        colors = ["#ffffff", "#efefef"]

        for idx, x in enumerate(self._results):
            w = SearchItemWidget()
            w.setFromResult(x)

            item = QListWidgetItem(self.results_list)
            item.setSizeHint(w.sizeHint())

            n = idx % len(colors)
            item.setBackground(QColor(colors[n]))

            self.results_list.addItem(item)
            self.results_list.setItemWidget(item, w)

    def _show_first_page(self):
        self.results_page_input.setValue(1)

    def _show_last_page(self):
        n = self.results_page_input.maximum()
        self.results_page_input.setValue(n)

    def connect_results_list_signals(self):

        self.results_list.itemSelectionChanged.connect(partial(self._handle_result_select))

        self.results_page_input \
            .valueChanged.connect(partial(self.render_results_page))

        self.results_first_page_btn \
            .clicked.connect(partial(self._show_first_page))

        self.results_last_page_btn \
            .clicked.connect(partial(self._show_last_page))



import logging as _logging
import shutil
from typing import Callable, List, Optional, TypedDict
import re

from whoosh.highlight import SCORE, HtmlFormatter
from whoosh.index import FileIndex, create_in, open_dir
from whoosh.fields import SchemaClass, NUMERIC, TEXT, ID
from whoosh.qparser import QueryParser, FuzzyTermPlugin
from whoosh.searching import Results, Hit
from whoosh.analysis import CharsetFilter, StemmingAnalyzer
from whoosh.support.charset import accent_map

from simsapa.app.helpers import compactPlainText, compactRichText
from simsapa.app.db import appdata_models as Am
from simsapa.app.db import userdata_models as Um
from simsapa import INDEX_DIR

logger = _logging.getLogger(__name__)

# Add an accent-folding filter to the stemming analyzer
folding_analyzer = StemmingAnalyzer() | CharsetFilter(accent_map)

class SuttasIndexSchema(SchemaClass):
    db_id = NUMERIC(stored = True, unique = True)
    schema_name = TEXT(stored = True)
    uid = ID(stored = True)
    title = TEXT(stored = True, analyzer = folding_analyzer)
    title_pali = TEXT(stored = True, analyzer = folding_analyzer)
    title_trans = TEXT(stored = True, analyzer = folding_analyzer)
    content = TEXT(stored = True, analyzer = folding_analyzer)
    ref = ID(stored = True)

class DictWordsIndexSchema(SchemaClass):
    db_id = NUMERIC(stored = True, unique = True)
    schema_name = TEXT(stored = True)
    uid = ID(stored = True)
    word = TEXT(stored = True, analyzer = folding_analyzer)
    synonyms = TEXT(stored = True, analyzer = folding_analyzer)
    content = TEXT(stored = True, analyzer = folding_analyzer)

class SearchResult(TypedDict):
    # database id
    db_id: int
    # database schema name (appdata or userdata)
    schema_name: str
    # database table name (e.g. suttas or dict_words)
    table_name: str
    uid: Optional[str]
    title: str
    ref: Optional[str]
    author: Optional[str]
    # highlighted snippet
    snippet: str
    # page number in a document
    page_number: Optional[int]

def sutta_hit_to_search_result(x: Hit, snippet: str) -> SearchResult:

    return SearchResult(
        db_id = x['db_id'],
        schema_name = x['schema_name'],
        table_name = 'suttas',
        uid = x['uid'],
        title = x['title'],
        ref = x['ref'],
        author = None,
        snippet = snippet,
        page_number = None,
    )

def dict_word_hit_to_search_result(x: Hit, snippet: str) -> SearchResult:
    return SearchResult(
        db_id = x['db_id'],
        schema_name = x['schema_name'],
        table_name = 'dict_words',
        uid = x['uid'],
        title = x['word'],
        ref = None,
        author = None,
        snippet = snippet,
        page_number = None,
    )

class SearchQuery:
    def __init__(self, ix: FileIndex, page_len: int, hit_to_result_fn: Callable):
        self.ix = ix
        self.searcher = ix.searcher()
        self.results: Results
        self.hits: int = 0
        self.page_len = page_len
        self.hit_to_result_fn = hit_to_result_fn

    def _compact_plain_snippet(self,
                               content: str,
                               title: Optional[str] = None,
                               ref: Optional[str] = None) -> str:

        s = content

        # AN 3.119 Kammantasutta
        # as added to the content when indexing
        if ref is not None and title is not None:
            s = s.replace(f"{ref} {title}", '')

        # 163–182. (- dash and -- en-dash)
        s = re.sub(r'[0-9\.–-]+', '', s)

        # ... Book of the Sixes 5.123.
        s = re.sub(r'Book of the [\w ]+[0-9\.]+', '', s)

        # Connected Discourses on ... 12.55.
        s = re.sub(r'\w+ Discourses on [\w ]+[0-9\.]+', '', s)

        # ...vagga
        s = re.sub(r'[\w -]+vagga', '', s)

        # ... Nikāya 123.
        s = re.sub(r'[\w]+ Nikāya +[0-9\.]*', '', s)

        # SC 1, (link text)
        s = re.sub('SC [0-9]+', '', s)

        # Remove the title from the content, but only the first instance, so as
        # not to remove a common word (e.g.'kamma') from the entire text.
        if title is not None:
            s = s.replace(title, '', 1)

        return s

    def _oneline(self, content: str) -> str:
        s = content
        # Clean up whitespace so that all text is one line
        s = s.replace("\n", ' ')
        # replace multiple spaces to one
        s = re.sub(r'  +', ' ', s)

        return s

    def _plain_snippet_from_hit(self, x: Hit) -> str:
            s = x['content'][0:500]
            if 'title' in x.keys() and 'ref' in x.keys():
                s = self._compact_plain_snippet(s, x['title'], x['ref'])
            else:
                s = self._compact_plain_snippet(s)
            s = self._oneline(s).strip()
            snippet = s[0:250] + ' ...'

            return snippet

    def _result_with_snippet_highlight(self, x: Hit):
        style = "<style>span.match { background-color: yellow; }</style>"
        fragments = x.highlights(fieldname='content', top=5)

        if len(fragments) > 0:
            snippet = style + fragments
        else:
            snippet = self._plain_snippet_from_hit(x)

        return self.hit_to_result_fn(x, snippet)

    def _result_with_content_no_highlight(self, x: Hit):
        snippet = self._plain_snippet_from_hit(x)
        return self.hit_to_result_fn(x, snippet)

    def _search_field(self, field_name: str, query: str) -> Results:
        parser = QueryParser(field_name, self.ix.schema)
        parser.add_plugin(FuzzyTermPlugin())
        q = parser.parse(query)

        # NOTE: limit=None is fast enough (~1700 hits for 'dhamma'), doesn't
        # hang when doing incremental search while typing.
        #
        # The highlighting _is_ slow, that has to be only done on the
        # displayed results page.

        return self.searcher.search(q, limit=None, terms=True)

    def highlight_results_page(self, page_num: int) -> List[SearchResult]:
        page_start = page_num * self.page_len
        page_end = page_start + self.page_len
        return list(map(self._result_with_snippet_highlight, self.results[page_start:page_end]))

    def new_query(self, query: str) -> List[SearchResult]:
        self.results = self._search_field('content', query)
        # NOTE: r.estimated_min_length() errors on some searches
        self.hits = len(self.results)

        if self.hits == 0:
            return []

        # may slow down highlighting with many results
        # r.fragmenter.charlimit = None

        self.results.fragmenter.maxchars = 200
        self.results.fragmenter.surround = 20
        # FIRST (the default) Show fragments in the order they appear in the document.
        # SCORE Show highest scoring fragments first.
        self.results.order = SCORE
        self.results.formatter = HtmlFormatter(tagname='span', classname='match')

        # NOTE: highlighting the matched fragments is slow, so only do
        # highlighting on the first page of results.

        return self.highlight_results_page(0)

    def all_results(self, highlight: bool = False) -> List[SearchResult]:
        if highlight:
            return list(map(self._result_with_snippet_highlight, self.results))
        else:
            return list(map(self._result_with_content_no_highlight, self.results))


class SearchIndexed:
    suttas_index: FileIndex
    dict_words_index: FileIndex

    def __init__(self):
        self.open_all()

    def index_all(self, db_session, only_if_empty: bool = False):
        if (not only_if_empty) or (only_if_empty and self.suttas_index.is_empty()):
            print("Indexing suttas ...")
            self.index_suttas(db_session, 'appdata')
            self.index_suttas(db_session, 'userdata')

        if (not only_if_empty) or (only_if_empty and self.dict_words_index.is_empty()):
            print("Indexing dict_words ...")
            self.index_dict_words(db_session, 'appdata')
            self.index_dict_words(db_session, 'userdata')

    def open_all(self):
        self.suttas_index: FileIndex = self._open_or_create_index('suttas', SuttasIndexSchema)
        self.dict_words_index: FileIndex = self._open_or_create_index('dict_words', DictWordsIndexSchema)

    def create_all(self, remove_if_exists: bool = True):
        if remove_if_exists and INDEX_DIR.exists():
            shutil.rmtree(INDEX_DIR)
        self.open_all()

    def _open_or_create_index(self, index_name: str, index_schema: SchemaClass) -> FileIndex:
        if not INDEX_DIR.exists():
            INDEX_DIR.mkdir(exist_ok=True)

        try:
            ix = open_dir(dirname = INDEX_DIR, indexname = index_name, schema = index_schema)
            return ix
        except Exception as e:
            logger.info(f"Can't open the index: {index_name}")
            print(e)

        try:
            ix = create_in(dirname = INDEX_DIR, indexname = index_name, schema = index_schema)
        except Exception as e:
            logger.error(f"Can't create the index: {index_name}")
            print(e)
            exit(1)

        return ix

    def index_suttas(self, db_session, schema_name: str):
        ix = self.suttas_index

        if schema_name == 'appdata':
            a = db_session.query(Am.Sutta).all()
        else:
            a = db_session.query(Um.Sutta).all()

        try:
            # NOTE: Only use multisegment=True when indexing from scratch.
            # Memory limit applies to each process individually.
            writer = ix.writer(procs=4, limitmb=256, multisegment=True)

            for i in a:
                # Prefer the html content field if not empty.
                if i.content_html is not None and len(i.content_html.strip()) > 0:
                    content = compactRichText(i.content_html)
                elif i.content_plain is None:
                    continue
                else:
                    content = compactPlainText(i.content_plain)

                # Add title and title_pali to content field so a single field query will match
                # Db fields can be None
                c = list(filter(lambda x: x is not None, [i.sutta_ref, i.title, i.title_pali]))
                pre = " ".join(c)
                if len(pre) > 0:
                    content = f"{pre} {content}"

                writer.add_document(
                    db_id = i.id,
                    schema_name = schema_name,
                    uid = i.uid,
                    title = i.title,
                    title_pali = i.title_pali,
                    title_trans = i.title_trans,
                    content = content,
                    ref = i.sutta_ref,
                )
            writer.commit()

        except Exception as e:
            logger.error("Can't index.")
            print(e)

    def index_dict_words(self, db_session, schema_name: str):
        ix = self.dict_words_index

        if schema_name == 'appdata':
            a = db_session.query(Am.DictWord).all()
        else:
            a = db_session.query(Um.DictWord).all()

        try:
            # NOTE: Only use multisegment=True when indexing from scratch.
            # Memory limit applies to each process individually.
            writer = ix.writer(procs=4, limitmb=256, multisegment=True)
            for i in a:
                # Prefer the html content field if not empty
                if i.definition_html is not None and len(i.definition_html.strip()) > 0:
                    content = compactRichText(i.definition_html)
                elif i.definition_plain is None:
                    continue
                else:
                    content = compactPlainText(i.definition_plain)

                # Add word and synonyms to content field so a single query will match
                if i.word is not None:
                    content = f"{i.word} {content}"

                if i.synonyms is not None:
                    content = f"{content} {i.synonyms}"

                writer.add_document(
                    db_id = i.id,
                    schema_name = schema_name,
                    uid = i.uid,
                    word = i.word,
                    synonyms = i.synonyms,
                    content = content,
                )
            writer.commit()

        except Exception as e:
            logger.error("Can't index.")
            print(e)




"""Baseline

Revision ID: 5aaa36ccfa37
Revises: 
Create Date: 2021-12-21 14:39:34.115783

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '5aaa36ccfa37'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    pass


def downgrade():
    pass



import logging as _logging
import re
import cgi
import socket
from http.server import HTTPServer, BaseHTTPRequestHandler

from simsapa import APP_QUEUES

logger = _logging.getLogger(__name__)


class Handler(BaseHTTPRequestHandler):
    def do_GET(self):
        self.send_response(404, 'Not Found')
        self._send_cors_headers()
        self.end_headers()

        # self.send_response(200)
        # self._send_cors_headers()
        # self.send_header('Content-type', 'text/html')
        # self.end_headers()
        # res = "<h1>It's over 9000!</h1>"
        # self.wfile.write(bytes(res, 'utf-8'))

    def do_OPTIONS(self):
        self.send_response(200)
        self._send_cors_headers()
        self.end_headers()

    def do_POST(self):
        path = self.path.rstrip('/')

        if re.search('/queues/*', path):

            ctype, pdict = cgi.parse_header(self.headers.get('content-type'))
            if ctype == 'application/json':
                length = int(self.headers.get('content-length'))
                data: str = self.rfile.read(length).decode('utf8')

            else:
                self.send_response(400, "Bad Request")
                self._send_cors_headers()
                self.end_headers()
                return

            queue_id = path.split('/')[-1]

            if queue_id not in APP_QUEUES.keys():
                self.send_response(403, 'Forbidden')
                self._send_cors_headers()
                self.end_headers()
                return

            APP_QUEUES[queue_id].put_nowait(data)

            self.send_response(200)
            self._send_cors_headers()
            self.end_headers()

        else:
            self.send_response(404, 'Not Found')
            self._send_cors_headers()
            self.end_headers()

    def _send_cors_headers(self):
        """ Sets headers required for CORS """
        self.send_header("Access-Control-Allow-Origin", "*")
        self.send_header("Access-Control-Allow-Methods", "GET,POST,OPTIONS")
        self.send_header("Access-Control-Allow-Headers", "x-api-key,Content-Type")


def start_server(port=8000):
    logger.info(f'Starting server on port {port}')
    httpd = HTTPServer(('127.0.0.1', port), Handler)
    httpd.serve_forever()


def find_available_port() -> int:
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.bind(('', 0))
    _, port = sock.getsockname()
    return port



from typing import List, Tuple, TypedDict, Optional
from pathlib import Path
from itertools import chain
from bokeh.document.locking import UnlockedDocumentProxy
from bokeh.models.renderers import GraphRenderer, GlyphRenderer

import networkx as nx
from bokeh.io import output_file, save, curdoc
from bokeh.document import Document
from bokeh.models import (Div, Button, Circle, MultiLine, Range1d, ColumnDataSource,
                          LabelSet, HoverTool, NodesAndLinkedEdges,
                          EdgesAndLinkedNodes)
from bokeh.palettes import Spectral8
from bokeh.plotting import figure, from_networkx, Figure
from bokeh.layouts import column
from bokeh.models import CustomJS
from bokeh import events

from .db import appdata_models as Am
from .db import userdata_models as Um

from .types import AppData, UDictWord, USutta, UDocument


class NodeData(TypedDict):
    label: str
    title: str
    description: str
    table: str
    id: int
    fill_color: str


GraphNode = Tuple[str, NodeData]

GraphEdge = Tuple[str, str]


class DocumentPage(TypedDict):
    doc: UDocument
    page_number: Optional[int]


def sutta_graph_id(x: USutta) -> str:
    schema = x.metadata.schema
    return f"{schema}.suttas.{x.id}"


def dict_word_graph_id(x: UDictWord) -> str:
    schema = x.metadata.schema
    return f"{schema}.dict_words.{x.id}"


def document_graph_id(x: DocumentPage) -> str:
    doc = x['doc']
    page = x['page_number']
    schema = doc.metadata.schema
    if page is None:
        return f"{schema}.documents.{doc.id}"
    else:
        return f"{schema}.documents.{doc.id}.{page}"


def sutta_to_node(x: USutta) -> GraphNode:
    schema = x.metadata.schema
    return (
        sutta_graph_id(x),
        NodeData(
            label = x.sutta_ref, # type: ignore
            title = x.title, # type: ignore
            description = '',
            table = f'{schema}.suttas',
            id = x.id, # type: ignore
            fill_color = Spectral8[0],
        )
    )


def dict_word_to_node(x: UDictWord) -> GraphNode:
    schema = x.metadata.schema
    return (
        dict_word_graph_id(x),
        NodeData(
            label = x.word, # type: ignore
            title = x.word, # type: ignore
            description = '',
            table = f'{schema}.dict_words',
            id = x.id, # type: ignore
            fill_color = Spectral8[1],
        )
    )


def document_page_to_node(x: DocumentPage) -> GraphNode:
    doc = x['doc']
    page_number = x['page_number']
    schema = doc.metadata.schema
    if page_number is None:
        label = doc.title
    else:
        label = f"{doc.title} (p.{page_number})"
    return (
        document_graph_id(x),
        NodeData(
            label = label, # type: ignore
            title = doc.title, # type: ignore
            description = '',
            table = f'{schema}.documents',
            id = doc.id, # type: ignore
            fill_color = Spectral8[4],
        )
    )


def unique_nodes(nodes: List[GraphNode]) -> List[GraphNode]:
    listed = []
    unique_nodes = []
    for i in nodes:
        e = i[0]
        if e not in listed:
            listed.append(e)
            unique_nodes.append(i)

    unique_nodes.sort(key=lambda x: x[0])

    return unique_nodes


def unique_edges(edges: List[GraphEdge]) -> List[GraphEdge]:
    listed = []
    unique_edges = []
    for i in edges:
        e = f"{i[0]},{i[1]}"
        if e not in listed:
            listed.append(e)
            unique_edges.append(i)

    unique_edges.sort(key=lambda x: f"{x[0]},{x[1]}")

    return unique_edges


def sutta_nodes_and_edges(app_data: AppData, sutta: USutta, distance: int = 1):
    links = []

    # TODO: Assuming all links are in userdata, made between
    # 'appdata.suttas' records

    r = app_data.db_session \
        .query(Um.Link.to_id) \
        .filter(Um.Link.from_table == "appdata.suttas") \
        .filter(Um.Link.to_table == "appdata.suttas") \
        .filter(Um.Link.from_id == sutta.id) \
        .all()

    links.extend(r)

    r = app_data.db_session \
        .query(Um.Link.from_id) \
        .filter(Um.Link.from_table == "appdata.suttas") \
        .filter(Um.Link.to_table == "appdata.suttas") \
        .filter(Um.Link.to_id == sutta.id) \
        .all()

    links.extend(r)

    # IDs without the current sutta ID
    ids = filter(lambda x: x != sutta.id, map(lambda x: x[0], links))
    # set() will contain unique items
    sutta_ids = list(set(ids))

    suttas = app_data.db_session \
        .query(Am.Sutta) \
        .filter(Am.Sutta.id.in_(sutta_ids)) \
        .all()

    nodes = list(map(sutta_to_node, suttas))

    def to_edge(x: USutta):
        from_id = sutta_graph_id(sutta)
        to_id = sutta_graph_id(x)
        if to_id < from_id:
            return (to_id, from_id)
        else:
            return (from_id, to_id)

    edges = list(map(to_edge, suttas))

    # Collect links from other nodes

    if distance > 1:
        for i in suttas:
            (n, e) = sutta_nodes_and_edges(app_data=app_data, sutta=i, distance=distance - 1)
            nodes.extend(n)
            edges.extend(e)

    # Append the current sutta as a node

    nodes.append(sutta_to_node(sutta))

    return (unique_nodes(nodes), unique_edges(edges))


def dict_word_nodes_and_edges(app_data: AppData, dict_word: UDictWord, distance: int = 1):
    schema = dict_word.metadata.schema

    links = []

    r = app_data.db_session \
        .query(Um.Link.to_id) \
        .filter(Um.Link.from_table == f"{schema}.dict_words") \
        .filter(Um.Link.to_table == "appdata.suttas") \
        .filter(Um.Link.from_id == dict_word.id) \
        .all()

    links.extend(r)

    r = app_data.db_session \
        .query(Um.Link.from_id) \
        .filter(Um.Link.from_table == f"{schema}.dict_words") \
        .filter(Um.Link.to_table == "appdata.suttas") \
        .filter(Um.Link.to_id == dict_word.id) \
        .all()

    links.extend(r)

    ids = map(lambda x: x[0], links)
    # set() will contain unique items
    sutta_ids = list(set(ids))

    suttas = app_data.db_session \
        .query(Am.Sutta) \
        .filter(Am.Sutta.id.in_(sutta_ids)) \
        .all()

    nodes = list(map(sutta_to_node, suttas))

    def to_edge(x: USutta):
        from_id = dict_word_graph_id(dict_word)
        to_id = sutta_graph_id(x)
        return (to_id, from_id)

    edges = list(map(to_edge, suttas))

    # Collect links from other nodes

    if distance > 1:
        for i in suttas:
            (n, e) = sutta_nodes_and_edges(app_data=app_data, sutta=i, distance=distance - 1)
            nodes.extend(n)
            edges.extend(e)

    # Append the current doument as a node

    nodes.append(dict_word_to_node(dict_word))

    return (unique_nodes(nodes), unique_edges(edges))


def document_page_nodes_and_edges(app_data: AppData, db_doc: Am.Document, page_number: int, distance: int = 1):
    links = app_data.db_session \
        .query(Um.Link.to_id) \
        .filter(Um.Link.from_table == "userdata.documents") \
        .filter(Um.Link.from_page_number == page_number) \
        .filter(Um.Link.to_table == "appdata.suttas") \
        .filter(Um.Link.from_id == db_doc.id) \
        .all()

    ids = map(lambda x: x[0], links)
    # set() will contain unique items
    sutta_ids = list(set(ids))

    suttas = app_data.db_session \
        .query(Am.Sutta) \
        .filter(Am.Sutta.id.in_(sutta_ids)) \
        .all()

    nodes = list(map(sutta_to_node, suttas))

    def to_edge(x: USutta):
        d = DocumentPage(doc=db_doc, page_number=page_number)
        from_id = document_graph_id(d)
        to_id = sutta_graph_id(x)
        return (to_id, from_id)

    edges = list(map(to_edge, suttas))

    # Collect links from other nodes

    if distance > 1:
        for i in suttas:
            (n, e) = sutta_nodes_and_edges(app_data=app_data, sutta=i, distance=distance - 1)
            nodes.extend(n)
            edges.extend(e)

    # Append the current doument as a node

    d = DocumentPage(doc=db_doc, page_number=page_number)
    nodes.append(document_page_to_node(d))

    return (unique_nodes(nodes), unique_edges(edges))


def _suttas_from_links(app_data: AppData, links: List[Um.Link]) -> List[USutta]:
    results: List[USutta] = []

    def to_appdata_sutta_ids(x: Um.Link):
        ids = []
        if x.from_table == 'appdata.suttas':
            ids.append(x.from_id)
        if x.to_table == 'appdata.suttas':
            ids.append(x.to_id)
        return ids

    a = list(map(to_appdata_sutta_ids, links))
    db_ids = list(set(chain.from_iterable(a)))

    r = app_data.db_session \
        .query(Am.Sutta) \
        .filter(Am.Sutta.id.in_(db_ids)) \
        .all()

    results.extend(r)

    return results


def _dict_words_from_links(app_data: AppData, links: List[Um.Link]) -> List[UDictWord]:
    results: List[UDictWord] = []

    def to_appdata_word_ids(x: Um.Link):
        ids = []
        if x.from_table == 'appdata.dict_words':
            ids.append(x.from_id)
        if x.to_table == 'appdata.dict_words':
            ids.append(x.to_id)
        return ids

    def to_userdata_word_ids(x: Um.Link):
        ids = []
        if x.from_table == 'userdata.dict_words':
            ids.append(x.from_id)
        if x.to_table == 'userdata.dict_words':
            ids.append(x.to_id)
        return ids

    a = list(map(to_appdata_word_ids, links))
    appdata_db_ids = list(set(chain.from_iterable(a)))

    a = list(map(to_userdata_word_ids, links))
    userdata_db_ids = list(set(chain.from_iterable(a)))

    r = app_data.db_session \
        .query(Am.DictWord) \
        .filter(Am.DictWord.id.in_(appdata_db_ids)) \
        .all()

    results.extend(r)

    r = app_data.db_session \
        .query(Um.DictWord) \
        .filter(Um.DictWord.id.in_(userdata_db_ids)) \
        .all()

    results.extend(r)

    return results


def _documents_and_pages_from_links(app_data: AppData, links: List[Um.Link]) -> List[DocumentPage]:
    results: List[DocumentPage] = []

    def to_appdata_items(x: Um.Link):
        items = []
        if x.from_table == 'appdata.documents':
            items.append((x.from_id, x.from_page_number))
        if x.to_table == 'appdata.documents':
            items.append((x.to_id, x.to_page_number))
        return items

    def to_userdata_items(x: Um.Link):
        items = []
        if x.from_table == 'userdata.documents':
            items.append((x.from_id, x.from_page_number))
        if x.to_table == 'userdata.documents':
            items.append((x.to_id, x.to_page_number))
        return items

    a = list(map(to_appdata_items, links))
    appdata_items = list(set(chain.from_iterable(a)))
    appdata_ids = list(map(lambda x: x[0], appdata_items))

    a = list(map(to_userdata_items, links))
    userdata_items = list(set(chain.from_iterable(a)))
    userdata_ids = list(map(lambda x: x[0], userdata_items))

    appdata_db_res = app_data.db_session \
        .query(Am.Document) \
        .filter(Am.Document.id.in_(appdata_ids)) \
        .all()

    userdata_db_res = app_data.db_session \
        .query(Um.Document) \
        .filter(Um.Document.id.in_(userdata_ids)) \
        .all()

    def id_to_doc(x, db_res):
        id = x[0]
        page = x[1]
        doc = list(filter(lambda i: i.id == id, db_res)) # type: ignore
        return DocumentPage(doc=doc[0], page_number=page)

    appdata_docpages = list(map(lambda x: id_to_doc(x, appdata_db_res), appdata_items))
    results.extend(appdata_docpages)

    userdata_docpages = list(map(lambda x: id_to_doc(x, userdata_db_res), userdata_items))
    results.extend(userdata_docpages)

    return results


def all_nodes_and_edges(app_data: AppData):

    links = app_data.db_session.query(Um.Link).all()

    suttas = _suttas_from_links(app_data, links)
    words = _dict_words_from_links(app_data, links)
    documents_and_pages = _documents_and_pages_from_links(app_data, links)

    nodes = []
    nodes.extend(list(map(sutta_to_node, suttas)))
    nodes.extend(list(map(dict_word_to_node, words)))
    nodes.extend(list(map(document_page_to_node, documents_and_pages)))

    def from_agrees(x, link, postfix):
        from_schema = link.from_table.replace(postfix, '')
        return x.metadata.schema == from_schema and x.id == link.from_id

    def from_doc_page_agrees(x, link, postfix):
        from_schema = link.from_table.replace(postfix, '')
        return \
            x['doc'].metadata.schema == from_schema and x['doc'].id == link.from_id and \
            x['page_number'] == link.from_page_number

    def to_agrees(x, link, postfix):
        to_schema = link.to_table.replace(postfix, '')
        return x.metadata.schema == to_schema and x.id == link.to_id

    def to_doc_page_agrees(x, link, postfix):
        to_schema = link.to_table.replace(postfix, '')
        return \
            x['doc'].metadata.schema == to_schema and x['doc'].id == link.to_id and \
            x['page_number'] == link.to_page_number

    def to_edge(link: Um.Link):
        if link.from_table.endswith('.suttas'):
            sutta = list(filter(lambda x: from_agrees(x, link, '.suttas'), suttas))
            if len(sutta) == 1:
                from_edge_id = sutta_graph_id(sutta[0])
            else:
                from_edge_id = ''

        elif link.from_table.endswith('.dict_words'):
            word = list(filter(lambda x: from_agrees(x, link, '.dict_words'), words))
            if len(word) == 1:
                from_edge_id = dict_word_graph_id(word[0])
            else:
                from_edge_id = ''

        elif link.from_table.endswith('.documents'):
            doc = list(filter(lambda x: from_doc_page_agrees(x, link, '.documents'), documents_and_pages))
            if len(doc) == 1:
                from_edge_id = document_graph_id(doc[0])
            else:
                from_edge_id = ''

        else:
            from_edge_id = ''

        if link.to_table.endswith('.suttas'):
            sutta = list(filter(lambda x: to_agrees(x, link, '.suttas'), suttas))
            if len(sutta) == 1:
                to_edge_id = sutta_graph_id(sutta[0])
            else:
                to_edge_id = ''

        elif link.to_table.endswith('.dict_words'):
            word = list(filter(lambda x: to_agrees(x, link, '.dict_words'), words))
            if len(word) == 1:
                to_edge_id = dict_word_graph_id(word[0])
            else:
                to_edge_id = ''

        elif link.to_table.endswith('.documents'):
            doc = list(filter(lambda x: to_doc_page_agrees(x, link, '.documents'), documents_and_pages))
            if len(doc) == 1:
                to_edge_id = document_graph_id(doc[0])
            else:
                to_edge_id = ''

        else:
            to_edge_id = ''

        if to_edge_id < from_edge_id:
            return (to_edge_id, from_edge_id)
        else:
            return (from_edge_id, to_edge_id)

    edges = list(filter(lambda x: x[0] != '' and x[1] != '', map(to_edge, links)))

    return (unique_nodes(nodes), unique_edges(edges))


def generate_graph(nodes, edges, selected_indices: List[int], queue_id: str, output_path: Path, messages_url: str):
    if len(nodes) == 0:
        return

    G = nx.Graph()
    G.add_nodes_from(nodes)
    G.add_edges_from(edges)

    plot: Figure = figure(
        title=None,
        plot_width=700,
        plot_height=400,
        x_range=Range1d(-1.1, 1.1),
        y_range=Range1d(-1.1, 1.1),
        tools="pan,tap,wheel_zoom,reset",
        active_scroll="wheel_zoom",
    )

    plot.xgrid.grid_line_color = None
    plot.ygrid.grid_line_color = None
    plot.xaxis.axis_line_color = None
    plot.yaxis.axis_line_color = None

    plot.xaxis.axis_label_text_color = None
    plot.yaxis.axis_label_text_color = None

    plot.xaxis.major_label_text_color = None
    plot.yaxis.major_label_text_color = None

    plot.xaxis.major_tick_line_color = None
    plot.yaxis.major_tick_line_color = None
    plot.xaxis.minor_tick_line_color = None
    plot.yaxis.minor_tick_line_color = None

    network_graph: GraphRenderer = from_networkx(G, nx.spring_layout, scale=0.8, center=(0, 0), seed=100)

    selection_color = Spectral8[6]

    network_graph.node_renderer.glyph = Circle(size=15, fill_color='fill_color')
    network_graph.node_renderer.selection_glyph = Circle(size=15, fill_color=selection_color)
    network_graph.node_renderer.nonselection_glyph = Circle(size=15, fill_color='fill_color')

    network_graph.edge_renderer.glyph = MultiLine(line_color="black", line_alpha=0.8, line_width=1)
    network_graph.edge_renderer.selection_glyph = MultiLine(line_color=selection_color, line_width=2)

    network_graph.selection_policy = NodesAndLinkedEdges()
    network_graph.inspection_policy = EdgesAndLinkedNodes()

    plot.renderers.append(network_graph)

    x, y = zip(*network_graph.layout_provider.graph_layout.values())
    source = ColumnDataSource({
        'x': x,
        'y': y,
        'label': [nodes[i][1]['label'] for i in range(len(x))],
        'title': [nodes[i][1]['title'] for i in range(len(x))],
        'description': [nodes[i][1]['description'] for i in range(len(x))],
        'table': [nodes[i][1]['table'] for i in range(len(x))],
        'id': [nodes[i][1]['id'] for i in range(len(x))],
    })

    tooltips = [
        ('Title', '@title')
    ]

    cr = plot.circle(x='x', y='y', source=source, size=25,
                     fill_color=None, line_color=None,
                     hover_fill_color=selection_color,
                     hover_alpha=0.8, hover_line_color="white")

    plot.add_tools(HoverTool(tooltips=tooltips, renderers=[cr]))

    labels = LabelSet(
        x='x',
        y='y',
        text='label',
        source=source,
        background_fill_color='white',
        text_font_size='15px',
        background_fill_alpha=.7)

    plot.renderers.append(labels)

    network_graph.node_renderer.data_source.selected.indices = selected_indices

    network_graph.node_renderer.data_source.selected.js_on_change('indices', CustomJS(args=dict(source=source), code="""
window.selected_info = [];

var inds = cb_obj.indices;

var body = document.getElementsByTagName('body')[0];

var desc_div = document.getElementById('selected_descriptions');

if (desc_div == null || typeof desc_div == 'undefined') {
    desc_div = document.createElement('div');
    desc_div.setAttribute('id', 'selected_descriptions');
    body.appendChild(desc_div);
}

desc_div.innerHTML = '';
var desc_div_content = '';

var data = source.data;

for (var i=0; i<inds.length; i++) {
    var idx = inds[i];

    var table = data['table'][idx];
    var id = data['id'][idx];
    var title = data['title'][idx];
    var desc_content = data['description'][idx];

    window.selected_info.push({ table: table, id: id });

    desc_div_content += '<h1 class="title">' + title + '</h1><div class="description">' + desc_content + '</div>';
}

desc_div.innerHTML = desc_div_content;
"""))

    button = Button(label='Open Sutta')

    js_code = """
if (typeof window.selected_info === 'undefined') {
    window.selected_info = [];
}
if (window.selected_info.length > 0) {
    const params = {
        action: 'show_sutta',
        arg: window.selected_info[0],
    };
    const options = {
        method: 'POST',
        headers: {
          'Accept': 'application/json',
          'Content-Type': 'application/json'
        },
        body: JSON.stringify(params),
    };
    fetch('%s', options);
}
""" % (messages_url,)

    button.js_on_event(events.ButtonClick, CustomJS(code=js_code, args={}))

    text = """
<style>
    #selected_descriptions {
        font-family: Helvetica, Arial, sans-serif;
        padding: 0 30px;
    }

    h1.title {
        font-style: normal;
        font-weight: normal;
        font-size: 20px;
        line-height: 25px;
        letter-spacing: 0.5px;
        color: #1a1a1a;
    }

    .description {
        font-size: 16px;
        line-height: 20px;
    }
</style>"""

    header = Div(text=text)
    layout = column(header, button, plot)

    doc: Document = curdoc()
    doc.clear()
    doc.add_root(layout)

    output_file(filename=str(output_path), mode='absolute')
    save(doc)



import os
import shutil
import json
import semver
import logging as _logging
from functools import partial
from typing import List, Optional
from pathlib import Path
from zipfile import ZipFile
import pandas
from pyglossary import Glossary
from sqlalchemy.sql import func

from PyQt5.QtCore import QAbstractListModel, Qt
from PyQt5.QtGui import QFont, QIcon
from PyQt5.QtWidgets import (QLabel, QMainWindow,
                             QMessageBox, QInputDialog, QFileDialog, QLineEdit)

from simsapa import ASSETS_DIR
from simsapa.assets import icons_rc  # noqa: F401

from ..app.db_models import DictionarySource
from ..app.types import AppData, DictWord
from ..app.helpers import download_file
from ..assets.ui.dictionaries_manager_window_ui import Ui_DictionariesManagerWindow

logger = _logging.getLogger(__name__)
Glossary.init()


class DictionarySourceListModel(QAbstractListModel):
    def __init__(self, *args, dictionaries=None, **kwargs):
        super(DictionarySourceListModel, self).__init__(*args, **kwargs)
        self.dictionaries = dictionaries or []

    def data(self, index, role):
        if role == Qt.DisplayRole:
            return self.dictionaries[index.row()].title

        if role == Qt.DecorationRole:
            if self.dictionaries[index.row()].has_update:
                return QIcon(":sync")

    def rowCount(self, index):
        return len(self.dictionaries)


class DictionariesManagerWindow(QMainWindow, Ui_DictionariesManagerWindow):
    def __init__(self, app_data: AppData, parent=None) -> None:
        super().__init__(parent)
        self.setupUi(self)

        self._app_data: AppData = app_data

        self.model = DictionarySourceListModel(dictionaries=self._get_all_dictionary_sources())

        self.dictionary_sources_list.setModel(self.model)
        self.sel_model = self.dictionary_sources_list.selectionModel()

        self._ui_setup()
        self._connect_signals()

        self.statusbar.showMessage("Ready", 3000)

    def _ui_setup(self):
        self.status_msg = QLabel("")
        self.statusbar.addPermanentWidget(self.status_msg)

        self._show_dictionary_source_clear()

    def _get_all_dictionary_sources(self) -> List[DictionarySource]:
        return self._app_data.user_db_session.query(DictionarySource).all()

    def get_selected_dictionary_source(self) -> Optional[DictionarySource]:
        a = self.dictionary_sources_list.selectedIndexes()
        if not a:
            return None

        item = a[0]
        return self.model.dictionaries[item.row()]

    def remove_selected_dictionary_source(self):
        a = self.dictionary_sources_list.selectedIndexes()
        if not a:
            return None

        # Remove from model
        item = a[0]
        dict_id = self.model.dictionaries[item.row()].id

        del self.model.dictionaries[item.row()]
        self.model.layoutChanged.emit()
        self.dictionary_sources_list.clearSelection()
        self._show_dictionary_source_clear()

        # Remove from database

        db_item = self._app_data.user_db_session \
                                .query(DictionarySource) \
                                .filter(DictionarySource.id == dict_id) \
                                .first()
        self._app_data.user_db_session.delete(db_item)
        self._app_data.user_db_session.commit()

    def _handle_dictionary_source_select(self):
        doc = self.get_selected_dictionary_source()
        if doc:
            self._show_dictionary_source(doc)
        else:
            self._show_dictionary_source_clear()

    def _show_dictionary_source_clear(self):
        self.status_msg.clear()
        self.dict_creator.clear()
        self.dict_cover.clear()
        self.dict_message.clear()
        self.dict_title.clear()
        self.dict_version.clear()
        self.dict_description.clear()
        self.dict_update_button.hide()

    def _show_dictionary_source(self, dict_source: DictionarySource):
        self.status_msg.setText(dict_source.title)
        self.dict_creator.setText(dict_source.creator)
        self.dict_title.setText(dict_source.title)
        self.dict_version.setText(dict_source.version)
        self.dict_description.setText(dict_source.description)

        if dict_source.has_update:
            self.dict_message.setText("An updated version is available.")
            font = QFont()
            font.setItalic(True)
            self.dict_message.setFont(font)
            self.dict_update_button.show()
        else:
            self.dict_message.clear()
            self.dict_update_button.hide()

    def add_words_from_xlsx(self, xlsx_path: Path):
        xlsx_data_df = pandas.read_excel(xlsx_path, sheet_name='Words')
        words_data = xlsx_data_df.to_dict(orient='records')

        def to_word(data) -> DictWord:
            word = DictWord(data['word'])
            word.definition_md = data['definition_md']
            return word

        words: List[DictWord] = list(map(to_word, words_data))

        print(len(words))

    def add_words_from_stardict(self, ifo_path: Path):
        logger.info(f'add_words_from_stardict(): {ifo_path}')
        glos = Glossary()
        if not glos.read(ifo_path, format='Stardict', direct=True, progressbar=False):
            logger.error('Reading StarDict failed')

        words: List[DictWord] = []

        n = 0
        max_words = 100  # FIXME don't stop importing

        # TODO create dictionary for manager from Startdict info

        for entry in glos:
            if n >= max_words:
                break
            n += 1

            # handle entry.defiFormat
            # m is plaintext
            # h is html

            logger.info(entry.s_word)
            word = DictWord(entry.s_word)
            word.definition_md = entry.defi
            words.append(word)

        print(len(words))

    def process_file(self, p: Path):
        # MS Excel Spreadsheet
        if p.suffix == '.xlsx':
            self.add_words_from_xlsx(p)

        # Stardict format
        if p.suffix == '.ifo':
            self.add_words_from_stardict(p)

    def add_words_from_data_zip(self, zip_path: Path, extract_dir: Path):
        try:
            with ZipFile(zip_path, 'r') as z:
                z.extractall(extract_dir)
        except Exception as e:
            logger.error(e)
            return

        for p in extract_dir.glob("*"):
            if p.is_dir():
                for pp in p.glob("*"):
                    self.process_file(pp)
            else:
                self.process_file(p)

    def add_from_file(self, zip_path: Path):
        temp_dir: Path = ASSETS_DIR.joinpath('temp')
        if not temp_dir.exists():
            os.mkdir(temp_dir)

        self.add_words_from_data_zip(zip_path, temp_dir)

        shutil.rmtree(temp_dir)

        # TODO add dictionary entry to manager

    def add_from_url(self, json_url):
        temp_dir: Path = ASSETS_DIR.joinpath('temp')
        if not temp_dir.exists():
            os.mkdir(temp_dir)

        json_path = download_file(json_url, temp_dir)
        with open(json_path, 'r') as f:
            remote_info = json.loads(f.read())

        zip_path = download_file(remote_info['data_zip_url'], temp_dir)

        self.add_words_from_data_zip(zip_path, temp_dir)

        # Remove downloaded files
        shutil.rmtree(temp_dir)

        # Add DictionarySource

        item = self._app_data.user_db_session \
                             .query(DictionarySource) \
                             .filter(DictionarySource.info_json_url == json_url) \
                             .first()

        if item is None:
            logger.info(f"Add new: {remote_info['title']}")
            db_source = DictionarySource(
                title=remote_info['title'],
                creator=remote_info['creator'],
                description=remote_info['description'],
                contact_email=remote_info['contact_email'],
                version=remote_info['version'],
                has_update=False,
                info_json_url=json_url,
                data_zip_url=remote_info['data_zip_url'],
                created_at=func.now(),
            )
            try:
                self._app_data.user_db_session.add(db_source)
                self._app_data.user_db_session.commit()

                # Add to model
                self.model.dictionaries.append(db_source)

            except Exception as e:
                logger.error(e)

        else:
            logger.info(f"Update: {item.title}")

            values = {
                'title': remote_info['title'],
                'creator': remote_info['creator'],
                'description': remote_info['description'],
                'contact_email': remote_info['contact_email'],
                'version': remote_info['version'],
                'has_update': False,
                'info_json_url': json_url,
                'data_zip_url': remote_info['data_zip_url'],
                'updated_at': func.now()
            }
            try:
                self._app_data.user_db_session \
                              .query(DictionarySource) \
                              .filter(DictionarySource.id == item.id) \
                              .update(values)

                self._app_data.user_db_session.commit()

            except Exception as e:
                logger.error(e)

        self._handle_dictionary_source_select()
        self.model.layoutChanged.emit()

    def _add_from_url_dialog(self):
        url, ok = QInputDialog.getText(self,
                                       "Add from URL...",
                                       "URL of JSON info of the dictionary:",
                                       QLineEdit.Normal,
                                       "")
        if ok:
            self.add_from_url(url)

    def _add_from_file_dialog(self):
        file_path, _ = QFileDialog.getOpenFileName(
            None,
            "Open File...",
            "",
            "Zip Files (*.zip)")

        if len(file_path) != 0:
            self.add_from_file(file_path)

    def _remove_dictionary_source_dialog(self):
        dict_source = self.get_selected_dictionary_source()
        if not dict_source:
            return

        reply = QMessageBox.question(self,
                                     'Remove Dictionary Source...',
                                     'Remove this Dictionary source?',
                                     QMessageBox.Yes | QMessageBox.No,
                                     QMessageBox.No)
        if reply == QMessageBox.Yes:
            self.remove_selected_dictionary_source()

    def _check_updates(self):
        temp_dir: Path = ASSETS_DIR.joinpath('temp')
        if not temp_dir.exists():
            os.mkdir(temp_dir)

        for d in self.model.dictionaries:
            json_path = download_file(d.info_json_url, temp_dir)
            with open(json_path, 'r') as f:
                remote_info = json.loads(f.read())
                if semver.compare(remote_info['version'], d.version) == 1:

                    d.has_update = True
                    self._app_data.user_db_session.commit()

        self.model.layoutChanged.emit()

    def _update_selected_dictionary_source(self):
        dict_source = self.get_selected_dictionary_source()
        if dict_source:
            self.add_from_url(dict_source.info_json_url)

    def _connect_signals(self):
        self.action_Close_Window \
            .triggered.connect(partial(self.close))

        self.action_Add_from_URL \
            .triggered.connect(partial(self._add_from_url_dialog))

        self.action_Import_from_File \
            .triggered.connect(partial(self._add_from_file_dialog))

        self.action_Check_Updates \
            .triggered.connect(partial(self._check_updates))

        self.action_Remove \
            .triggered.connect(partial(self._remove_dictionary_source_dialog))

        self.sel_model.selectionChanged.connect(partial(self._handle_dictionary_source_select))

        self.dict_update_button.clicked.connect(partial(self._update_selected_dictionary_source))




def html_page(content: str, messages_url: str):
    css = "pre { white-space: pre-wrap; }"

    js = """
document.addEventListener('DOMContentLoaded', function() {
    links = document.getElementsByTagName('a');
    for (var i=0; i<links.length; i++) {
        links[i].onclick = function(e) {
            url = e.target.href;
            if (!url.startsWith('sutta:') && !url.startsWith('word:')) {
                return;
            }

            e.preventDefault();

            var params = {};

            if (url.startsWith('sutta:')) {
                s = url.replace('sutta:', '');
                params = {
                    action: 'show_sutta_by_uid',
                    arg: {'uid': s},
                };
            } else if (url.startsWith('word:')) {
                s = url.replace('word:', '');
                params = {
                    action: 'show_word_by_uid',
                    arg: {'uid': s},
                };
            }
            const options = {
                method: 'POST',
                headers: {
                  'Accept': 'application/json',
                  'Content-Type': 'application/json'
                },
                body: JSON.stringify(params),
            };
            fetch('%s', options);
        }
    }
});
""" % (messages_url,)

    page_html = """
<!doctype html>
<html>
<head>
    <meta charset="utf-8">
    <style>%s</style>
    <script>%s</script>
</head>
<body>
%s
</body>
</html>
""" % (css, js, content)

    return page_html
